#!/usr/bin/env python

import socket
import threading
import sys
import os
import curses
import socks
import stem.process
from wh import WHMgr
import random
from binascii import a2b_base64 as a2b
from binascii import b2a_base64 as b2a
from getpass import getpass
from smp import SMP
from stem.control import Controller
from stem.util import term
from curses.textpad import Textbox
from contextlib import contextmanager
from pyaxo import Axolotl, hash_
from time import sleep

"""
Standalone chat script using libsodium for encryption with the Axolotl
ratchet for key management.

This version of the chat client makes connections over the tor network.
The server creates an ephemeral hidden service and the client connects
to the hidden service. You will need to load the following additional
python modules for this to work: stem, pysocks, txtorcon, pysocks,
and magic-wormhole. They are available on pypi via pip.

Axotor also requires tor (>=2.9.1). Currently this is in the unstable
branch. So you may need to update your distribution's tor repository
accordingly.

The only inputs required are the nicks of the conversants, as well as
a master key. The master key must be of the form N-x where N is an
integer (1-3 digits should be fine) and x is a lower-case alphanumeric
string. The initial axolotl database configuration and credential
exchange are derived from the master key. The master key should be
exchanged out-of-band between the two conversants before communication
can be established. An example master key might be 293-xyzzy (don't use
this one).

On start, the program initializes a tor server and exchanges axolotl
and hidden service authentication credentials over tor using PAKE
(password-authenticated key exchange). The specific implementation of
PAKE used is https://github.com/warner/magic-wormhole. The server
side creates an ephemeral hidden service that requires basic_auth
to connect.

The hidden service and axolotl credentials are ephemeral. They exist
only in ram, and will be deleted upon exit from the program. Exit by
typing .quit at the chat prompt.

The client also does an authentication step using the Socialist
Millionaire's Protocol. During startup, after a network connection is
established, you will be prompted for a secret. If the secret
matches that input by the other party, a chat is established and
the input window text appears in green. If the secret does not match
the other party's secret, you will be prompted whether or not to
continue. If you continue, the input window text will appear in red
to remind you that the session is unauthenticated.

The Axolotl protocol is actually authenticated through the key
agreement process when the credentials are created. You may wonder why
the additional SMP authentication step is included. The answer lies in
the fact that between sessions, the key databases are stored on disk
and - at least in principle - could be tampered with. This SMP step
assures that the other party to your session is actually who you think
it is.

Usage:
1. One side starts the server with:
     axotor.py -s

2. The other side connects the client to the server with:
     axotor.py -c

3. .quit at the chat prompt will quit (don't forget the "dot")

4. .send <filename> will send a file to the other party. The file
   can be from anywhere in the filesystem on the sending computer
   (~/a/b/<filename> supported) and will be stored in the receiver's
   local directory.

Axochat requires the Axolotl module at https://github.com/rxcomm/pyaxo

Copyright (C) 2015-2016 by David R. Andersen <k0rx@RXcomm.net>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

TOR_SERVER_PORT             = 9054
TOR_SERVER_CONTROL_PORT     = 9055
TOR_CLIENT_PORT             = 9154
TOR_CLIENT_CONTROL_PORT     = 9155
CLIENT_FILE_TX_PORT         = 2000
SERVER_FILE_TX_PORT         = 2001

# An attempt to limit the damage from this bug in curses:
# https://bugs.python.org/issue13051
# The input textbox is 8 rows high. So assuming a maximum
# terminal width of 512 columns, we arrive at 8x512=4096.
# Most terminal windows should be smaller than this.
sys.setrecursionlimit(4096)

@contextmanager
def socketcontext(*args, **kwargs):
    s = socket.socket(*args, **kwargs)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    yield s
    s.close()

@contextmanager
def torcontext():
    try:
        s = socks.socksocket()
        s.set_proxy(socks.SOCKS5, '127.0.0.1', TOR_CLIENT_PORT)
        yield s
        s.close()
    except socks.SOCKS5Error:
        print ''
        print 'You need to wait long enough for the Hidden Service'
        print 'at the server to be established. Try again in a'
        print 'minute or two.'

class _Textbox(Textbox):
    """
    curses.textpad.Textbox requires users to ^g on completion, which is sort
    of annoying for an interactive chat client such as this, which typically only
    reuquires an enter. This subclass fixes this problem by signalling completion
    on Enter as well as ^g. Also, map <Backspace> key to ^h.
    """
    def __init__(*args, **kwargs):
        Textbox.__init__(*args, **kwargs)

    def do_command(self, ch):
        if ch == curses.KEY_RESIZE:
            resizeWindows()
            for i in range(8): # delete 8 input window lines
                Textbox.do_command(self, 1)
                Textbox.do_command(self, 11)
                Textbox.do_command(self, 16)
            return Textbox.do_command(self, 7)
        if ch == 10: # Enter
            return 0
        if ch == 127: # Backspace
            return 8
        return Textbox.do_command(self, ch)

def validator(ch):
    """
    Update screen if necessary and release the lock so receiveThread can run
    """
    global screen_needs_update
    try:
        if screen_needs_update:
            curses.doupdate()
            screen_needs_update = False
        return ch
    finally:
        winlock.release()
        sleep(0.01) # let receiveThread in if necessary
        winlock.acquire()

def windowFactory():
    stdscr = curses.initscr()
    curses.noecho()
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(1, curses.COLOR_RED, -1)
    curses.init_pair(2, curses.COLOR_GREEN, -1)
    curses.init_pair(3, curses.COLOR_YELLOW, -1)
    curses.cbreak()
    curses.curs_set(1)
    (sizey, sizex) = stdscr.getmaxyx()
    input_win = curses.newwin(8, sizex, sizey-8, 0)
    output_win = curses.newwin(sizey-8, sizex, 0, 0)
    input_win.idlok(1)
    input_win.scrollok(1)
    input_win.nodelay(1)
    input_win.leaveok(0)
    input_win.timeout(100)
    output_win.idlok(1)
    output_win.scrollok(1)
    output_win.leaveok(0)
    return stdscr, input_win, output_win

def closeWindows(stdscr):
    curses.nocbreak()
    stdscr.keypad(0)
    curses.echo()
    curses.endwin()

def resizeWindows():
    global stdscr, input_win, output_win, textpad, text_color
    temp_win = output_win
    yold, xold = output_win.getmaxyx()
    stdscr, input_win, output_win = windowFactory()
    stdscr.noutrefresh()
    ynew, xnew = output_win.getmaxyx()
    if yold > ynew:
        sminrow = yold - ynew
        dminrow = 0
    else:
        sminrow = 0
        dminrow = ynew - yold
    temp_win.overwrite(output_win, sminrow, 0, dminrow, 0, ynew-1, xnew-1)
    del temp_win
    output_win.move(ynew-1, 0)
    output_win.noutrefresh()
    input_win.attron(text_color)
    input_win.noutrefresh()
    curses.doupdate()
    textpad = _Textbox(input_win, insert_mode=True)
    textpad.stripspaces = True

def usage():
    print 'Usage: ' + sys.argv[0] + ' -(s,c)'
    print ' -s: start a chat in server mode'
    print ' -c: start a chat in client mode'
    print 'quit with .quit'
    print 'send file with .send <filename>'
    sys.exit(1)

def reportTransferSocketError():
    global output_win
    with winlock:
        output_win.addstr('Socket error: Something went wrong.\n',
                           curses.color_pair(1))
        output_win.refresh()
    sys.exit()

def sendFile(s, filename, abort):
    global axolotl, output_win
    if abort:
        with cryptlock:
            data = axolotl.encrypt('ABORT') + 'EOP'
        s.send(data)
        try:
            s.recv(3, socket.MSG_WAITALL)
        except socket.error:
            pass
        sys.exit()
    else:
        with winlock:
            output_win.addstr('Sending file %s...\n' % filename,
                               curses.color_pair(3))
            output_win.refresh()
    with open(filename, 'rb') as f:
        data = f.read()
        if len(data) == 0:
            data = 'Sender tried to send a null file!'
        with cryptlock:
            data = axolotl.encrypt(data)
        s.send(data + 'EOP')
    try:
        s.recv(3, socket.MSG_WAITALL)
    except socket.error:
        pass

def receiveFile(s, filename):
    global axolotl, output_win
    data = ''
    while data[-3:] != 'EOP':
        rcv = s.recv(4096)
        data = data + rcv
        if not rcv:
            with winlock:
                output_win.addstr('Receiving %s aborted...\n' % filename,
                                   curses.color_pair(1))
                output_win.refresh()
            try:
                s.send('EOP')
            except socket.error:
                pass
            sys.exit()
    with cryptlock:
        data = axolotl.decrypt(data[:-3])
    if data == 'ABORT':
        with winlock:
            output_win.addstr('Receiving %s aborted...\n' % filename,
                               curses.color_pair(1))
            output_win.refresh()
        sys.exit()
    with open(filename, 'wb') as f:
        f.write(data)
    try:
        s.send('EOP')
    except socket.error:
        pass

def uploadThread(onion, command):
    global output_win
    with transferlock:
        filename = command.split(':> .send ')[1].strip()
        filename = os.path.expanduser(filename)
        if not os.path.exists(filename) or os.path.isdir(filename):
            with winlock:
                output_win.addstr('File %s does not exist...\n' % filename,
                                   curses.color_pair(1))
                output_win.refresh()
            abort = True
        else:
            abort = False
        if onion is None:
            with socketcontext(socket.AF_INET, socket.SOCK_STREAM) as s:
                try:
                    s.bind(('localhost', SERVER_FILE_TX_PORT))
                    s.listen(1)
                    conn, addr = s.accept()
                except socket.error:
                    reportTransferSocketError()
                sendFile(conn, filename, abort)
        else:
            with torcontext() as s:
                try:
                    s.connect((onion, CLIENT_FILE_TX_PORT))
                except socket.error:
                    reportTransferSocketError()
                sendFile(s, filename, abort)
        with winlock:
            output_win.addstr('%s sent...\n' % filename, curses.color_pair(3))
            output_win.refresh()

def downloadThread(onion, command):
    global output_win
    with transferlock:
        filename = command.split(':> .send ')[1].strip().split('/').pop()
        with winlock:
            output_win.addstr('Receiving file %s...\n' % filename,
                               curses.color_pair(3))
            output_win.refresh()
        if onion is None:
            with socketcontext(socket.AF_INET, socket.SOCK_STREAM) as s:
                try:
                    s.bind(('localhost', CLIENT_FILE_TX_PORT))
                    s.listen(1)
                    conn, addr = s.accept()
                except socket.error:
                    reportTransferSocketError()
                receiveFile(conn, filename)
        else:
            with torcontext() as s:
                try:
                    s.connect((onion, SERVER_FILE_TX_PORT))
                except socket.error:
                    reportTransferSocketError()
                receiveFile(s, filename)
        with winlock:
           output_win.addstr('%s received...\n' % filename, curses.color_pair(3))
           output_win.refresh()

def receiveThread(sock, text_color, onion):
    global screen_needs_update, a, stdscr, input_win, output_win
    while True:
        data = ''
        while data[-3:] != 'EOP':
            rcv = sock.recv(1024)
            if not rcv:
                input_win.move(0, 0)
                input_win.addstr('Disconnected - Ctrl-C to exit!',
                                 text_color)
                input_win.refresh()
                sys.exit()
            data = data + rcv
        data_list = data.split('EOP')
        with winlock:
            (cursory, cursorx) = input_win.getyx()
            for data in data_list:
                if data != '':
                    with cryptlock:
                        data = axolotl.decrypt(data)
                if ':> .quit' in data:
                    closeWindows(stdscr)
                    print 'The other party exited the chat...'
                    sleep(1.5)
                    os._exit(0)
                if ':> .send ' in data:
                    t = threading.Thread(target=downloadThread,
                                         args=(onion,data))
                    t.start()
                    data = ''
                output_win.addstr(data)
            input_win.move(cursory, cursorx)
            input_win.cursyncup()
            input_win.noutrefresh()
            output_win.noutrefresh()
            screen_needs_update = True

def chatThread(sock, smp_match, onion):
    global screen_needs_update, axolotl, stdscr, input_win, \
           output_win, textpad, text_color
    stdscr, input_win, output_win = windowFactory()
    y, x = output_win.getmaxyx()
    output_win.move(y-1, 0)
    if smp_match:
        text_color = curses.color_pair(2) # green
    else:
        text_color = curses.color_pair(1) # red
    input_win.attron(text_color)
    input_win.addstr(0, 0, NICK + ':> ')
    textpad = _Textbox(input_win, insert_mode=True)
    textpad.stripspaces = True
    t = threading.Thread(target=receiveThread,
                         args=(sock, text_color, onion))
    t.daemon = True
    t.start()
    try:
        while True:
            with winlock:
                data = textpad.edit(validator)
                if len(data) != 0 and chr(127) not in data:
                    input_win.clear()
                    input_win.addstr(NICK+':> ')
                    output_win.addstr(data.replace('\n', '') + '\n', text_color)
                    output_win.noutrefresh()
                    input_win.move(0, len(NICK)+3)
                    input_win.cursyncup()
                    input_win.noutrefresh()
                    screen_needs_update = True
                    data = data.replace('\n', '') + '\n'
                    try:
                        with cryptlock:
                            sock.send(axolotl.encrypt(data) + 'EOP')
                    except socket.error:
                        input_win.addstr('Disconnected')
                        input_win.refresh()
                        closeWindows(stdscr)
                        sys.exit()
                    if NICK+':> .quit' in data:
                        closeWindows(stdscr)
                        print 'Notifying the other party that you are quitting...'
                        sys.exit()
                    elif NICK+':> .send ' in data:
                        t = threading.Thread(target=uploadThread,
                                             args=(onion,data))
                        t.start()
                else:
                    input_win.addstr(NICK+':> ')
                    input_win.move(0, len(NICK)+3)
                    input_win.cursyncup()
                    input_win.noutrefresh()
                    screen_needs_update = True
            if screen_needs_update:
                curses.doupdate()
                screen_needs_update = False
    except KeyboardInterrupt:
        closeWindows(stdscr)

def tor(port, controlport, tor_dir, descriptor_cookie):
    tor_process = stem.process.launch_tor_with_config(
        tor_cmd = 'tor',
        config = {
                  'ControlPort': str(controlport),
                  'SocksPort'  : str(port),
                  'Log'        : [ 'NOTICE stdout', 'ERR file /tmp/tor_error_log', ],
                  'DataDirectory' : tor_dir,
                  'CookieAuthentication': '1',
                  'HidServAuth': descriptor_cookie,
                 },
        completion_percent = 100,
        take_ownership = True,
        timeout = 90,
        init_msg_handler = print_bootstrap_lines,
        )
    return tor_process

def print_bootstrap_lines(line):
    if 'Bootstrapped ' in line:
        print(term.format(line, term.Color.RED))

def clientController(descriptor_cookie, onion):
    controller = Controller.from_port(address='127.0.0.1',
                                      port=TOR_CLIENT_CONTROL_PORT)
    try:
        controller.authenticate()
        controller.set_options([
            ('HidServAuth', onion + ' ' + descriptor_cookie),
            ])
    except Exception as e:
        print e
    return controller

def ephemeralHiddenService():
    PORT = 50000
    HOST = '127.0.0.1'
    print 'Waiting for Hidden Service descriptor to be published...'
    print ' (this may take some time)'

    controller = Controller.from_port(address='127.0.0.1',
                                      port=TOR_SERVER_CONTROL_PORT)
    controller.authenticate()
    try:
        hs = controller.create_ephemeral_hidden_service([PORT,
                                                        CLIENT_FILE_TX_PORT,
                                                        SERVER_FILE_TX_PORT],
                                                        basic_auth={'axotor': None},
                                                        await_publication=True)
    except Exception as e:
        print e
    return controller, hs.client_auth['axotor'], hs.service_id +'.onion'

def credentialsSend(mkey, cookie, ratchet_key, onion):
    w = WHMgr(unicode(mkey),
              cookie+'___'+ratchet_key+'___'+onion,
              u'tcp:127.0.0.1:'+unicode(TOR_SERVER_CONTROL_PORT))
    w.send()
    w.run()
    return w.confirmed

def credentialsReceive(mkey):
    w = WHMgr(unicode(mkey), None, u'tcp:127.0.0.1:'+unicode(TOR_CLIENT_CONTROL_PORT))
    w.receive()
    w.run()
    return w.data

def smptest(secret, sock, is_server):
    global axolotl
    # Create an SMP object with the calculated secret
    smp = SMP(secret)

    if is_server:
        # Do the SMP protocol
        buffer = sock.recv(2435, socket.MSG_WAITALL)
        padlength = ord(buffer[-1:])
        buffer = axolotl.decrypt(buffer[:-padlength])
        buffer = smp.step2(buffer)
        buffer = axolotl.encrypt(buffer)
        padlength = 4535-len(buffer)
        buffer = buffer+padlength*chr(padlength) # pad to fixed length
        sock.send(buffer)

        buffer = sock.recv(3465, socket.MSG_WAITALL)
        padlength = ord(buffer[-1:])
        buffer = axolotl.decrypt(buffer[:-padlength])
        buffer = smp.step4(buffer)
        buffer = axolotl.encrypt(buffer)
        padlength = 1365-len(buffer)
        buffer = buffer+padlength*chr(padlength) # pad to fixed length
        sock.send(buffer)

    else:
        # Do the SMP protocol
        buffer = smp.step1()
        buffer = axolotl.encrypt(buffer)
        padlength = 2435-len(buffer)
        buffer = buffer+padlength*chr(padlength) # pad to fixed length
        sock.send(buffer)

        buffer = sock.recv(4535, socket.MSG_WAITALL)
        padlength = ord(buffer[-1:])
        buffer = axolotl.decrypt(buffer[:-padlength])
        buffer = smp.step3(buffer)
        buffer = axolotl.encrypt(buffer)
        padlength = 3465-len(buffer)
        buffer = buffer+padlength*chr(padlength) # pad to fixed length
        sock.send(buffer)

        buffer = sock.recv(1365, socket.MSG_WAITALL)
        padlength = ord(buffer[-1:])
        buffer = axolotl.decrypt(buffer[:-padlength])
        smp.step5(buffer)

    # Check if the secrets match
    if smp.match:
        print 'Secrets Match!'
        sleep(1)
        smp_match = True
    else:
        print 'Secrets DO NOT Match!'
        smp_match = False
    return smp_match

def doSMP(sock, is_server):
    global axolotl
    ans = raw_input('Run SMP authentication step? (y/N)? ')
    if not ans == 'y': ans = 'N'
    sock.send(axolotl.encrypt(ans))
    data = sock.recv(121, socket.MSG_WAITALL)
    data = axolotl.decrypt(data)
    if ans == 'N' and data == 'y':
        print 'Other party requested SMP authentication'
    if ans == 'y' or data == 'y':
        print 'Performing per-session SMP authentication...'
        ans = getpass('Enter SMP secret: ')
        print 'Running SMP protocol...'
        secret = ans + axolotl.state['CONVid']
        smp_match = smptest(secret, sock, is_server)
        if not smp_match:
            ans = raw_input('Continue? (y/N) ')
            if ans != 'y':
                print 'Exiting...'
                sys.exit()
    else:
        print 'OK - skipping SMP step and assuming ' + \
              'the other party is already authenticated...'
        smp_match = True
        sleep(2)
    return smp_match

if __name__ == '__main__':
    global axolotl
    try:
        mode = sys.argv[1]
    except:
        usage()

    NICK = raw_input('Enter your nick: ')
    OTHER_NICK = 'x'
    winlock = threading.Lock()
    transferlock = threading.Lock()
    cryptlock = threading.Lock()
    screen_needs_update = False
    HOST = '127.0.0.1'
    PORT=50000
    mkey = getpass('What is the masterkey (format: NNN-xxxx)? ')

    if mode == '-s':
        axolotl = Axolotl(NICK,
                    dbname=OTHER_NICK+'.db',
                    dbpassphrase=None,
                    nonthreaded_sql=False)
        axolotl.createState(other_name=OTHER_NICK,
                           mkey=hash_(mkey),
                           mode=False)
        tor_process = tor(TOR_SERVER_PORT,
                          TOR_SERVER_CONTROL_PORT,
                          '/tmp/tor.server',
                          '')
        hs, cookie, onion = ephemeralHiddenService()
        print 'Exchanging credentials via tor...'
        if credentialsSend(mkey,
                           cookie,
                           b2a(axolotl.state['DHRs']).strip(),
                           onion):
            pass
        else:
            sys.exit(1)
        print 'Credentials sent, waiting for the other party to connect...'
        with socketcontext(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((HOST, PORT))
            s.listen(1)
            conn, addr = s.accept()
            print 'Connected...'
            smp_match = doSMP(conn, True)
            chatThread(conn, smp_match, None)

    elif mode == '-c':
        axolotl = Axolotl(NICK,
                    dbname=OTHER_NICK+'.db',
                    dbpassphrase=None,
                    nonthreaded_sql=False)
        tor_process = tor(TOR_CLIENT_PORT,
                          TOR_CLIENT_CONTROL_PORT,
                          '/tmp/tor.client',
                          '')
        print 'Exchanging credentials via tor...'
        creds = credentialsReceive(mkey)
        if not creds:
            print 'Master Key Mismatch!'
            print 'Exiting...'
            sys.exit()
        cookie, rkey, onion = creds.split('___')
        controller = clientController(cookie, onion)
        axolotl.createState(other_name=OTHER_NICK,
                      mkey=hash_(mkey),
                      mode=True,
                      other_ratchetKey=a2b(rkey))

        print 'Credentials received, connecting to the other party...'
        with torcontext() as s:
            s.connect((onion, PORT))
            print 'Connected...'
            smp_match = doSMP(s, False)
            chatThread(s, smp_match, onion)

    else:
        usage()



#!/usr/bin/env python

import hashlib
import socket
import threading
import sys
import curses
from curses.textpad import Textbox
from random import randint
from contextlib import contextmanager
from pyaxo import Axolotl
from time import sleep
from getpass import getpass
from binascii import a2b_base64 as a2b
from binascii import b2a_base64 as b2a

"""
Standalone chat script using libsodium for encryption with the Axolotl
ratchet for key management.

Usage:
1. One side starts the server with:
     axochat.py -s

2. The other side connects the client to the server with:
     axochat.py -c

3. Both sides need to input the same master key. This can be any
   alphanumeric string. Also, the server will generate a handshake
   key that is a required input for the client.

4. .quit at the chat prompt will quit (don't forget the "dot")

Port 50000 is the default port, but you can choose your own port as well.

Axochat requires the Axolotl module at https://github.com/rxcomm/pyaxo

Copyright (C) 2014-2016 by David R. Andersen <k0rx@RXcomm.net>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""

@contextmanager
def socketcontext(*args, **kwargs):
    s = socket.socket(*args, **kwargs)
    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
    yield s
    s.close()

class _Textbox(Textbox):
    """
    curses.textpad.Textbox requires users to ^g on completion, which is sort
    of annoying for an interactive chat client such as this, which typically only
    reuquires an enter. This subclass fixes this problem by signalling completion
    on Enter as well as ^g. Also, map <Backspace> key to ^h.
    """
    def __init__(*args, **kwargs):
        Textbox.__init__(*args, **kwargs)

    def do_command(self, ch):
        if ch == 10: # Enter
            return 0
        if ch == 127: # Enter
            return 8
        return Textbox.do_command(self, ch)

def validator(ch):
    """
    Update screen if necessary and release the lock so receiveThread can run
    """
    global screen_needs_update
    try:
        if screen_needs_update:
            curses.doupdate()
            screen_needs_update = False
        return ch
    finally:
        lock.release()
        sleep(0.01) # let receiveThread in if necessary
        lock.acquire()

def windows():
    stdscr = curses.initscr()
    curses.noecho()
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(3, 2, -1)
    curses.cbreak()
    curses.curs_set(1)
    (sizey, sizex) = stdscr.getmaxyx()
    input_win = curses.newwin(8, sizex, sizey-8, 0)
    output_win = curses.newwin(sizey-8, sizex, 0, 0)
    input_win.idlok(1)
    input_win.scrollok(1)
    input_win.nodelay(1)
    input_win.leaveok(0)
    input_win.timeout(100)
    input_win.attron(curses.color_pair(3))
    output_win.idlok(1)
    output_win.scrollok(1)
    output_win.leaveok(0)
    return stdscr, input_win, output_win

def closeWindows(stdscr):
    curses.nocbreak()
    stdscr.keypad(0)
    curses.echo()
    curses.endwin()

def usage():
    print 'Usage: ' + sys.argv[0] + ' -(s,c)'
    print ' -s: start a chat in server mode'
    print ' -c: start a chat in client mode'
    exit()

def receiveThread(sock, stdscr, input_win, output_win):
    global screen_needs_update, a
    while True:
        data = ''
        while data[-3:] != 'EOP':
            rcv = sock.recv(1024)
            if not rcv:
                input_win.move(0, 0)
                input_win.addstr('Disconnected - Ctrl-C to exit!')
                input_win.refresh()
                sys.exit()
            data = data + rcv
        data_list = data.split('EOP')
        lock.acquire()
        (cursory, cursorx) = input_win.getyx()
        for data in data_list:
            if data != '':
                output_win.addstr(a.decrypt(data))
        input_win.move(cursory, cursorx)
        input_win.cursyncup()
        input_win.noutrefresh()
        output_win.noutrefresh()
        sleep(0.01) # write time for axo db
        screen_needs_update = True
        lock.release()

def chatThread(sock):
    global screen_needs_update, a
    stdscr, input_win, output_win = windows()
    input_win.addstr(0, 0, NICK + ':> ')
    textpad = _Textbox(input_win, insert_mode=True)
    textpad.stripspaces = True
    t = threading.Thread(target=receiveThread, args=(sock, stdscr, input_win,output_win))
    t.daemon = True
    t.start()
    try:
        while True:
            lock.acquire()
            data = textpad.edit(validator)
            if NICK+':> .quit' in data:
                closeWindows(stdscr)
                sys.exit()
            input_win.clear()
            input_win.addstr(NICK+':> ')
            output_win.addstr(data.replace('\n', '') + '\n', curses.color_pair(3))
            output_win.noutrefresh()
            input_win.move(0, len(NICK)+3)
            input_win.cursyncup()
            input_win.noutrefresh()
            screen_needs_update = True
            data = data.replace('\n', '') + '\n'
            try:
                sock.send(a.encrypt(data) + 'EOP')
            except socket.error:
                input_win.addstr('Disconnected')
                input_win.refresh()
                closeWindows(stdscr)
                sys.exit()
            sleep(0.01) # write time for axo db
            lock.release()
    except KeyboardInterrupt:
        closeWindows(stdscr)

def getPasswd(nick):
    return '1'

if __name__ == '__main__':
    global a
    try:
        mode = sys.argv[1]
    except:
        usage()

    NICK = raw_input('Enter your nick: ')
    OTHER_NICK = raw_input('Enter the nick of the other party: ')
    mkey = getpass('Enter the master key: ')
    lock = threading.Lock()
    screen_needs_update = False
    HOST = ''
    while True:
        try:
            PORT = raw_input('TCP port (1 for random choice, 50000 is default): ')
            PORT = int(PORT)
            break
        except ValueError:
            PORT = 50000
            break
    if PORT >= 1025 and PORT <= 65535:
        pass
    elif PORT == 1:
        PORT = 1025 + randint(0, 64510)
        print 'PORT is ' + str(PORT)

    if mode == '-s':
        a = Axolotl(NICK,
                    dbname=OTHER_NICK+'.db',
                    dbpassphrase=None,
                    nonthreaded_sql=False)
        a.createState(other_name=OTHER_NICK,
                      mkey=hashlib.sha256(mkey).digest(),
                      mode=False)
        print 'Your ratchet key is: %s' % b2a(a.state['DHRs']).strip()
        print 'Send this to %s...' % OTHER_NICK

        print 'Waiting for ' + OTHER_NICK + ' to connect...'
        with socketcontext(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind((HOST, PORT))
            s.listen(1)
            conn, addr = s.accept()
            chatThread(conn)

    elif mode == '-c':
        rkey = raw_input('Enter %s\'s ratchet key: ' % OTHER_NICK)
        a = Axolotl(NICK,
                    dbname=OTHER_NICK+'.db',
                    dbpassphrase=None,
                    nonthreaded_sql=False)
        a.createState(other_name=OTHER_NICK,
                      mkey=hashlib.sha256(mkey).digest(),
                      mode=True,
                      other_ratchetKey=a2b(rkey))

        HOST = raw_input('Enter the server: ')
        print 'Connecting to ' + HOST + '...'
        with socketcontext(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((HOST, PORT))
            chatThread(s)

    else:
        usage()



import errno
import os


DEFAULT_DB = './axolotl.db'

PLAINTEXT = 'message {}'

EXCHANGE_IDS = list()
EXCHANGES = list()


def remove_db(file_path=DEFAULT_DB):
    try:
        os.remove(file_path)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise


def encrypt(axolotl, i, pt, ct):
    ct.append(axolotl.encrypt(pt[i]))


def decrypt(axolotl, i, pt, ct):
    assert axolotl.decrypt(ct[i]) == pt[i]


def exchange_0(a, b):
    pt = [PLAINTEXT.format(i) for i in range(14)]
    ct = list()

    encrypt(a, 0, pt, ct)
    encrypt(a, 1, pt, ct)
    encrypt(b, 2, pt, ct)
    decrypt(b, 0, pt, ct)
    decrypt(b, 1, pt, ct)
    decrypt(a, 2, pt, ct)
    encrypt(a, 3, pt, ct)
    encrypt(a, 4, pt, ct)
    encrypt(b, 5, pt, ct)
    encrypt(a, 6, pt, ct)
    encrypt(b, 7, pt, ct)
    encrypt(a, 8, pt, ct)
    encrypt(a, 9, pt, ct)
    encrypt(a, 10, pt, ct)
    encrypt(a, 11, pt, ct)
    decrypt(b, 11, pt, ct)
    decrypt(b, 3, pt, ct)
    decrypt(b, 9, pt, ct)
    decrypt(a, 5, pt, ct)
    decrypt(a, 7, pt, ct)
    decrypt(b, 4, pt, ct)
    encrypt(b, 12, pt, ct)
    decrypt(a, 12, pt, ct)
    encrypt(a, 13, pt, ct)
    decrypt(b, 13, pt, ct)
    decrypt(b, 6, pt, ct)


def exchange_1(a, b):
    n = 3
    pt = list()
    ct = list()

    for i in range(n):
        pt.append(PLAINTEXT.format(i))
        encrypt(a, i, pt, ct)

    for i in range(n):
        decrypt(b, i, pt, ct)

    for i in range(n, n*2):
        pt.append(PLAINTEXT.format(i))
        encrypt(b, i, pt, ct)

    for i in range(n, n*2):
        decrypt(a, i, pt, ct)


def exchange_2(a, b):
    n = 3
    pt = list()
    ct = list()

    for i in range(n):
        pt.append(PLAINTEXT.format(i))
        encrypt(a, i, pt, ct)

    for i in reversed(range(n)):
        decrypt(b, i, pt, ct)

    for i in range(n, n*2):
        pt.append(PLAINTEXT.format(i))
        encrypt(b, i, pt, ct)

    for i in reversed(range(n, n*2)):
        decrypt(a, i, pt, ct)


def exchange_3(a, b):
    pt = [PLAINTEXT.format(i) for i in range(6)]
    ct = list()

    encrypt(a, 0, pt, ct)
    decrypt(b, 0, pt, ct)
    encrypt(b, 1, pt, ct)
    decrypt(a, 1, pt, ct)
    encrypt(a, 2, pt, ct)
    encrypt(a, 3, pt, ct)
    decrypt(b, 2, pt, ct)
    encrypt(b, 4, pt, ct)
    decrypt(a, 4, pt, ct)
    encrypt(a, 5, pt, ct)
    decrypt(b, 5, pt, ct)
    decrypt(b, 3, pt, ct)


for i in range(4):
    id_ = 'exchange_{}'.format(i)
    EXCHANGE_IDS.append(id_)
    EXCHANGES.append(globals()[id_])



from collections import namedtuple
from Queue import Queue
from threading import Event, Thread

import pytest

from pyaxo import Axolotl


def test_init_conversation(axolotl_a, axolotl_b,
                           a_identity_keys, b_identity_keys,
                           a_handshake_keys, b_handshake_keys,
                           a_ratchet_keys, b_ratchet_keys,
                           exchange):
    conv_a = axolotl_a.init_conversation(
        axolotl_b.name,
        priv_identity_key=a_identity_keys.priv,
        identity_key=a_identity_keys.pub,
        priv_handshake_key=a_handshake_keys.priv,
        other_identity_key=b_identity_keys.pub,
        other_handshake_key=b_handshake_keys.pub,
        priv_ratchet_key=a_ratchet_keys.priv,
        ratchet_key=a_ratchet_keys.pub,
        other_ratchet_key=b_ratchet_keys.pub)

    conv_b = axolotl_b.init_conversation(
        axolotl_a.name,
        priv_identity_key=b_identity_keys.priv,
        identity_key=b_identity_keys.pub,
        priv_handshake_key=b_handshake_keys.priv,
        other_identity_key=a_identity_keys.pub,
        other_handshake_key=a_handshake_keys.pub,
        priv_ratchet_key=b_ratchet_keys.priv,
        ratchet_key=b_ratchet_keys.pub,
        other_ratchet_key=a_ratchet_keys.pub)

    exchange(conv_a, conv_b)


def test_init_nonthreaded_conversations(
        axolotl_a, axolotl_b, axolotl_c,
        a_identity_keys, b_identity_keys, c_identity_keys,
        a_handshake_keys, b_handshake_keys, c_handshake_keys,
        a_ratchet_keys, b_ratchet_keys, c_ratchet_keys,
        exchange):
    conversations = initialize_conversations(
        axolotl_a, a_identity_keys, a_handshake_keys, a_ratchet_keys,
        axolotl_b, b_identity_keys, b_handshake_keys, b_ratchet_keys,
        axolotl_c, c_identity_keys, c_handshake_keys, c_ratchet_keys)

    exchange(conversations.ab, conversations.ba)
    exchange(conversations.ac, conversations.ca)
    exchange(conversations.bc, conversations.cb)


def test_init_threaded_conversations(
        threaded_axolotl_a, threaded_axolotl_b, threaded_axolotl_c,
        a_identity_keys, b_identity_keys, c_identity_keys,
        a_handshake_keys, b_handshake_keys, c_handshake_keys,
        a_ratchet_keys, b_ratchet_keys, c_ratchet_keys,
        exchange):
    conversations = initialize_conversations(
        threaded_axolotl_a, a_identity_keys, a_handshake_keys, a_ratchet_keys,
        threaded_axolotl_b, b_identity_keys, b_handshake_keys, b_ratchet_keys,
        threaded_axolotl_c, c_identity_keys, c_handshake_keys, c_ratchet_keys)

    run_threaded_exchanges(exchange, conversations)


def test_create_conversation(axolotl_a, axolotl_b,
                             a_identity_keys, b_identity_keys,
                             a_handshake_keys, b_handshake_keys,
                             a_ratchet_keys, b_ratchet_keys,
                             exchange):
    mkey = 'masterkey'

    conv_a = axolotl_a.create_conversation(
        other_name=axolotl_b.name,
        mkey=mkey,
        mode=True,
        priv_identity_key=a_identity_keys.priv,
        identity_key=a_identity_keys.pub,
        other_identity_key=b_identity_keys.pub,
        other_ratchet_key=b_ratchet_keys.pub)

    conv_b = axolotl_b.create_conversation(
        other_name=axolotl_a.name,
        mkey=mkey,
        mode=False,
        priv_identity_key=b_identity_keys.priv,
        identity_key=b_identity_keys.pub,
        other_identity_key=a_identity_keys.pub,
        priv_ratchet_key=b_ratchet_keys.priv,
        ratchet_key=b_ratchet_keys.pub)

    exchange(conv_a, conv_b)


def test_create_nonthreaded_conversations(
        axolotl_a, axolotl_b, axolotl_c,
        a_identity_keys, b_identity_keys, c_identity_keys,
        a_handshake_keys, b_handshake_keys, c_handshake_keys,
        a_ratchet_keys, b_ratchet_keys, c_ratchet_keys,
        exchange):
    mkey_ab = 'masterkey_ab'
    mkey_ac = 'masterkey_ac'
    mkey_bc = 'masterkey_bc'

    conversations = create_conversations(
        axolotl_a, a_identity_keys, a_handshake_keys, a_ratchet_keys,
        axolotl_b, b_identity_keys, b_handshake_keys, b_ratchet_keys,
        axolotl_c, c_identity_keys, c_handshake_keys, c_ratchet_keys,
        mkey_ab, mkey_ac, mkey_bc)

    exchange(conversations.ab, conversations.ba)
    exchange(conversations.ac, conversations.ca)
    exchange(conversations.bc, conversations.cb)


def test_create_threaded_conversations(
        threaded_axolotl_a, threaded_axolotl_b, threaded_axolotl_c,
        a_identity_keys, b_identity_keys, c_identity_keys,
        a_handshake_keys, b_handshake_keys, c_handshake_keys,
        a_ratchet_keys, b_ratchet_keys, c_ratchet_keys,
        exchange):
    mkey_ab = 'masterkey_ab'
    mkey_ac = 'masterkey_ac'
    mkey_bc = 'masterkey_bc'

    conversations = create_conversations(
        threaded_axolotl_a, a_identity_keys, a_handshake_keys, a_ratchet_keys,
        threaded_axolotl_b, b_identity_keys, b_handshake_keys, b_ratchet_keys,
        threaded_axolotl_c, c_identity_keys, c_handshake_keys, c_ratchet_keys,
        mkey_ab, mkey_ac, mkey_bc)

    run_threaded_exchanges(exchange, conversations)


@pytest.fixture()
def threaded_axolotl_a():
    return Axolotl('Angie', dbpassphrase=None, nonthreaded_sql=False)


@pytest.fixture()
def threaded_axolotl_b():
    return Axolotl('Barb', dbpassphrase=None, nonthreaded_sql=False)


@pytest.fixture()
def threaded_axolotl_c():
    return Axolotl('Charlie', dbpassphrase=None, nonthreaded_sql=False)


class ThreadedExchange(Thread):
    def __init__(self, exchange, axolotl_x, axolotl_y, event, queue):
        super(ThreadedExchange, self).__init__()
        self.daemon = True
        self.exchange = exchange
        self.axolotl_x = axolotl_x
        self.axolotl_y = axolotl_y
        self.event = event
        self.queue = queue

    def run(self):
        self.event.wait()
        try:
            self.exchange(self.axolotl_x, self.axolotl_y)
        except AssertionError:
            self.queue.put(False)
        else:
            self.queue.put(True)


Conversations = namedtuple('Conversations', 'ab ac ba bc ca cb')


def initialize_conversations(
        axolotl_a, a_identity_keys, a_handshake_keys, a_ratchet_keys,
        axolotl_b, b_identity_keys, b_handshake_keys, b_ratchet_keys,
        axolotl_c, c_identity_keys, c_handshake_keys, c_ratchet_keys):
    ab = axolotl_a.init_conversation(
        other_name=axolotl_b.name,
        priv_identity_key=a_identity_keys.priv,
        identity_key=a_identity_keys.pub,
        priv_handshake_key=a_handshake_keys.priv,
        other_identity_key=b_identity_keys.pub,
        other_handshake_key=b_handshake_keys.pub,
        priv_ratchet_key=a_ratchet_keys.priv,
        ratchet_key=a_ratchet_keys.pub,
        other_ratchet_key=b_ratchet_keys.pub)

    ac = axolotl_a.init_conversation(
        other_name=axolotl_c.name,
        priv_identity_key=a_identity_keys.priv,
        identity_key=a_identity_keys.pub,
        priv_handshake_key=a_handshake_keys.priv,
        other_identity_key=c_identity_keys.pub,
        other_handshake_key=c_handshake_keys.pub,
        priv_ratchet_key=a_ratchet_keys.priv,
        ratchet_key=a_ratchet_keys.pub,
        other_ratchet_key=c_ratchet_keys.pub)

    ba = axolotl_b.init_conversation(
        other_name=axolotl_a.name,
        priv_identity_key=b_identity_keys.priv,
        identity_key=b_identity_keys.pub,
        priv_handshake_key=b_handshake_keys.priv,
        other_identity_key=a_identity_keys.pub,
        other_handshake_key=a_handshake_keys.pub,
        priv_ratchet_key=b_ratchet_keys.priv,
        ratchet_key=b_ratchet_keys.pub,
        other_ratchet_key=a_ratchet_keys.pub)

    bc = axolotl_b.init_conversation(
        other_name=axolotl_c.name,
        priv_identity_key=b_identity_keys.priv,
        identity_key=b_identity_keys.pub,
        priv_handshake_key=b_handshake_keys.priv,
        other_identity_key=c_identity_keys.pub,
        other_handshake_key=c_handshake_keys.pub,
        priv_ratchet_key=b_ratchet_keys.priv,
        ratchet_key=b_ratchet_keys.pub,
        other_ratchet_key=c_ratchet_keys.pub)

    ca = axolotl_c.init_conversation(
        other_name=axolotl_a.name,
        priv_identity_key=c_identity_keys.priv,
        identity_key=c_identity_keys.pub,
        priv_handshake_key=c_handshake_keys.priv,
        other_identity_key=a_identity_keys.pub,
        other_handshake_key=a_handshake_keys.pub,
        priv_ratchet_key=c_ratchet_keys.priv,
        ratchet_key=b_ratchet_keys.pub,
        other_ratchet_key=a_ratchet_keys.pub)

    cb = axolotl_c.init_conversation(
        other_name=axolotl_b.name,
        priv_identity_key=c_identity_keys.priv,
        identity_key=c_identity_keys.pub,
        priv_handshake_key=c_handshake_keys.priv,
        other_identity_key=b_identity_keys.pub,
        other_handshake_key=b_handshake_keys.pub,
        priv_ratchet_key=c_ratchet_keys.priv,
        ratchet_key=c_ratchet_keys.pub,
        other_ratchet_key=b_ratchet_keys.pub)

    return Conversations(ab, ac, ba, bc, ca, cb)


def create_conversations(
        axolotl_a, a_identity_keys, a_handshake_keys, a_ratchet_keys,
        axolotl_b, b_identity_keys, b_handshake_keys, b_ratchet_keys,
        axolotl_c, c_identity_keys, c_handshake_keys, c_ratchet_keys,
        mkey_ab, mkey_ac, mkey_bc):
    ab = axolotl_a.create_conversation(
        other_name=axolotl_b.name,
        mkey=mkey_ab,
        mode=True,
        priv_identity_key=a_identity_keys.priv,
        identity_key=a_identity_keys.pub,
        other_identity_key=b_identity_keys.pub,
        other_ratchet_key=b_ratchet_keys.pub)

    ac = axolotl_a.create_conversation(
        other_name=axolotl_c.name,
        mkey=mkey_ac,
        mode=False,
        priv_identity_key=a_identity_keys.priv,
        identity_key=a_identity_keys.pub,
        other_identity_key=c_identity_keys.pub,
        priv_ratchet_key=a_ratchet_keys.priv,
        ratchet_key=a_ratchet_keys.pub)

    ba = axolotl_b.create_conversation(
        other_name=axolotl_a.name,
        mkey=mkey_ab,
        mode=False,
        priv_identity_key=b_identity_keys.priv,
        identity_key=b_identity_keys.pub,
        other_identity_key=a_identity_keys.pub,
        priv_ratchet_key=b_ratchet_keys.priv,
        ratchet_key=b_ratchet_keys.pub)

    bc = axolotl_b.create_conversation(
        other_name=axolotl_c.name,
        mkey=mkey_bc,
        mode=True,
        priv_identity_key=b_identity_keys.priv,
        identity_key=b_identity_keys.pub,
        other_identity_key=c_identity_keys.pub,
        other_ratchet_key=c_ratchet_keys.pub)

    ca = axolotl_c.create_conversation(
        other_name=axolotl_a.name,
        mkey=mkey_ac,
        mode=True,
        priv_identity_key=c_identity_keys.priv,
        identity_key=c_identity_keys.pub,
        other_identity_key=a_identity_keys.pub,
        other_ratchet_key=a_ratchet_keys.pub)

    cb = axolotl_c.create_conversation(
        other_name=axolotl_b.name,
        mkey=mkey_bc,
        mode=False,
        priv_identity_key=c_identity_keys.priv,
        identity_key=c_identity_keys.pub,
        other_identity_key=b_identity_keys.pub,
        priv_ratchet_key=c_ratchet_keys.priv,
        ratchet_key=c_ratchet_keys.pub)

    return Conversations(ab, ac, ba, bc, ca, cb)


def run_threaded_exchanges(exchange, conversations):
    event = Event()

    queue_ab = Queue()
    exchange_ab = ThreadedExchange(exchange,
                                   conversations.ab, conversations.ba,
                                   event, queue_ab)

    queue_ac = Queue()
    exchange_ac = ThreadedExchange(exchange,
                                   conversations.ac, conversations.ca,
                                   event, queue_ac)

    queue_bc = Queue()
    exchange_bc = ThreadedExchange(exchange,
                                   conversations.bc, conversations.cb,
                                   event, queue_bc)

    exchange_ab.start()
    exchange_ac.start()
    exchange_bc.start()

    event.set()

    assert queue_ab.get() and queue_ac.get() and queue_bc.get()



import pytest

from pyaxo import Axolotl, generate_keypair

from . import utils


@pytest.fixture()
def a_identity_keys():
    return generate_keypair()


@pytest.fixture()
def b_identity_keys():
    return generate_keypair()


@pytest.fixture()
def c_identity_keys():
    return generate_keypair()


@pytest.fixture()
def a_handshake_keys():
    return generate_keypair()


@pytest.fixture()
def b_handshake_keys():
    return generate_keypair()


@pytest.fixture()
def c_handshake_keys():
    return generate_keypair()


@pytest.fixture()
def a_ratchet_keys():
    return generate_keypair()


@pytest.fixture()
def b_ratchet_keys():
    return generate_keypair()


@pytest.fixture()
def c_ratchet_keys():
    return generate_keypair()


@pytest.fixture()
def axolotl_a():
    return Axolotl('Angie', dbpassphrase=None)


@pytest.fixture()
def axolotl_b():
    return Axolotl('Barb', dbpassphrase=None)


@pytest.fixture()
def axolotl_c():
    return Axolotl('Charlie', dbpassphrase=None)


@pytest.fixture(params=utils.EXCHANGES, ids=utils.EXCHANGE_IDS)
def exchange(request):
    return request.param



#!/usr/bin/env python

"""
This script will add a new conversation between two parties to the
axolotl database.

You will need to provide the other party's name, identity key, handshake key,
and ratchet key. Conversations are identified by your name and the other person's
name. Each conversation should have a unique other person's name. Your name can be
the same for each conversation or different for each one or any combination.

If you decide not to complete the initialization process, just answer no to the
question about creating a new conversation. Nothing will be saved.

If you want to reinitialize a conversation, just run the script again.
The old conversation key data will be overwritten in the database.
"""

import sys
import binascii
from pyaxo import Axolotl

your_name = raw_input('Your name for this conversation? ').strip()
a = Axolotl(your_name)
a.printKeys()

ans = raw_input('Do you want to create a new conversation? y/N ').strip()
if ans == 'y':
    other_name = raw_input('What is the name of the other party? ').strip()
    identity = raw_input('What is the identity key for the other party? ').strip()
    handshake = raw_input('What is the handshake key for the other party? ').strip()
    ratchet = raw_input('What is the ratchet key for the other party? ').strip()
    a.initState(other_name, binascii.a2b_base64(identity), binascii.a2b_base64(handshake),
                binascii.a2b_base64(ratchet))
    a.saveState()
    print 'The conversation ' + your_name + ' -> ' + other_name + ' has been saved.'
else:
    print 'OK, nothing has been saved...'




#!/usr/bin/env python2
import os
import hashlib as h
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from wormhole.cli.public_relay import RENDEZVOUS_RELAY
from wormhole.wormhole import wormhole
from wormhole.tor_manager import TorManager
from wormhole.timing import DebugTiming

class WHMgr(object):
    def __init__(self, code, data, tor_port):
        self._code = code
        self._tor_port = tor_port
        self._reactor = reactor
        self._timing = DebugTiming()
        self.confirmed = False
        self.data = data
        self._tmgr = TorManager(self._reactor,
                                False,
                                self._tor_port,
                                timing=self._timing)

    @inlineCallbacks
    def send(self):
        """I send data through a wormhole and return True/False
        depending on whether or not the hash matched at the receive
        side.
        """
        def _confirm(input_message):
            if input_message == 'Confirmed!':
                self.confirmed = True

        if not self._tmgr.tor_available():
            print 'tor not available'
            os._exit(1)
        yield self._tmgr.start()
        self._w = wormhole(u'axotor', RENDEZVOUS_RELAY, self._reactor,
                           self._tmgr, timing=self._timing)
        self._w.set_code(self._code)
        self._w.send(self.data+h.sha256(self.data).hexdigest())
        self._d = self._w.get()
        self._d.addCallback(_confirm)
        self._d.addCallback(lambda _: self._w.close())
        self._d.addCallback(lambda _: self._reactor.stop())
        yield self._d
        return

    @inlineCallbacks
    def receive(self):
        """I receive data+hash, check for a match, confirm or not
        confirm to the sender, and return the data payload.
        """
        def _receive(inbound_message):
            self.data = inbound_message[:-64]
            _hash = inbound_message[-64:]
            if h.sha256(self.data).hexdigest() == _hash:
                self._w.send('Confirmed!')
            else:
                self._w.send('Not Confirmed!')

        if not self._tmgr.tor_available():
            print 'tor not available'
            os._exit(1)
        yield self._tmgr.start()
        self._w = wormhole(u"axotor", RENDEZVOUS_RELAY, self._reactor,
                           self._tmgr, timing=self._timing)
        self._w.set_code(self._code)
        self._d = self._w.get()
        self._d.addCallback(_receive)
        self._d.addCallback(lambda _: self._w.close())
        self._d.addCallback(lambda _: self._reactor.stop())
        yield self._d
        return

    def run(self):
        self._reactor.run()



import pytest

from pyaxo import generate_keypair


@pytest.fixture()
def keypair():
    return generate_keypair()


def test_keypair_tuple(keypair):
    assert isinstance(keypair, tuple)
    assert keypair.priv == keypair[0] and keypair.pub == keypair[1]


def test_keypair_different_values(keypair):
    assert keypair.priv != keypair.pub


def test_keypair_bytes(keypair):
    assert isinstance(keypair.priv, bytes) and isinstance(keypair.pub, bytes)



"""
pyaxo.py - a python implementation of the axolotl ratchet protocol.
https://github.com/trevp/axolotl/wiki/newversion

Symmetric encryption is done using the python-gnupg module.

Copyright (C) 2014 by David R. Andersen <k0rx@RXcomm.net>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

For more information, see https://github.com/rxcomm/pyaxo
"""

import errno
import os
import sqlite3
import sys
from collections import namedtuple
from getpass import getpass
from threading import Lock
from time import time

import nacl.secret
import nacl.utils
from nacl.encoding import Base64Encoder
from nacl.exceptions import CryptoError
from nacl.hash import sha256
from nacl.public import PrivateKey, PublicKey, Box
from passlib.utils.pbkdf2 import pbkdf2


ALICE_MODE = True
BOB_MODE = False

SALTS = {'RK': b'\x00',
         'HK': {ALICE_MODE: b'\x01', BOB_MODE: b'\x02'},
         'NHK': {ALICE_MODE: b'\x03', BOB_MODE: b'\x04'},
         'CK': {ALICE_MODE: b'\x05', BOB_MODE: b'\x06'},
         'CONVid': b'\x07'}

HEADER_LEN = 80
HEADER_PAD_NUM_LEN = 1
HEADER_COUNT_NUM_LEN = 3

class Axolotl(object):

    def __init__(self, name, dbname='axolotl.db', dbpassphrase='', nonthreaded_sql=True):
        self.name = name
        self.dbname = dbname
        self.nonthreaded_sql = nonthreaded_sql
        if dbpassphrase is None:
            self.dbpassphrase = None
        elif dbpassphrase != '':
            self.dbpassphrase = hash_(dbpassphrase)
        else:
            self.dbpassphrase = getpass('Database passphrase for '+ self.name + ': ').strip()
        self.conversation = AxolotlConversation(self, keys=dict(), mode=None)
        self.state['DHIs_priv'], self.state['DHIs'] = generate_keypair()
        self.state['DHRs_priv'], self.state['DHRs'] = generate_keypair()
        self.handshakeKey, self.handshakePKey = generate_keypair()
        self.storeTime = 2*86400 # minimum time (seconds) to store missed ephemeral message keys
        self.lock = Lock()
        self.persistence = SqlitePersistence(self.dbname,
                                             self.dbpassphrase,
                                             self.storeTime,
                                             self.nonthreaded_sql)

    @property
    def state(self):
        return self.conversation.keys

    @state.setter
    def state(self, state):
        self.conversation.keys = state

    @property
    def mode(self):
        return self.conversation.mode

    @mode.setter
    def mode(self, mode):
        self.conversation.mode = mode

    @property
    def db(self):
        return self.persistence.db

    @db.setter
    def db(self, db):
        self.persistence.db = db

    def tripleDH(self, a, a0, B, B0):
        if self.mode == None:
            sys.exit(1)
        return generate_3dh(a, a0, B, B0, self.mode)

    def genDH(self, a, B):
        return generate_dh(a, B)

    def genKey(self):
        return generate_keypair()

    def initState(self, other_name, other_identityKey, other_handshakeKey,
                  other_ratchetKey, verify=True):
        if verify:
            print 'Confirm ' + other_name + ' has identity key fingerprint:\n'
            fingerprint = hash_(other_identityKey).encode('hex').upper()
            fprint = ''
            for i in range(0, len(fingerprint), 4):
                fprint += fingerprint[i:i+2] + ':'
            print fprint[:-1] + '\n'
            print 'Be sure to verify this fingerprint with ' + other_name + \
                  ' by some out-of-band method!'
            print 'Otherwise, you may be subject to a Man-in-the-middle attack!\n'
            ans = raw_input('Confirm? y/N: ').strip()
            if ans != 'y':
                print 'Key fingerprint not confirmed - exiting...'
                sys.exit()

        self.conversation = self.init_conversation(other_name,
                                                   self.state['DHIs_priv'],
                                                   self.state['DHIs'],
                                                   self.handshakeKey,
                                                   other_identityKey,
                                                   other_handshakeKey,
                                                   self.state['DHRs_priv'],
                                                   self.state['DHRs'],
                                                   other_ratchetKey)

    def init_conversation(self, other_name,
                          priv_identity_key, identity_key, priv_handshake_key,
                          other_identity_key, other_handshake_key,
                          priv_ratchet_key=None, ratchet_key=None,
                          other_ratchet_key=None, mode=None):
        if mode is None:
            if identity_key < other_identity_key:
                mode = ALICE_MODE
            else:
                mode = BOB_MODE

        mkey = generate_3dh(priv_identity_key, priv_handshake_key,
                            other_identity_key, other_handshake_key,
                            mode)

        return self.create_conversation(other_name,
                                        mkey,
                                        mode,
                                        priv_identity_key,
                                        identity_key,
                                        other_identity_key,
                                        priv_ratchet_key,
                                        ratchet_key,
                                        other_ratchet_key)

    def createState(self, other_name, mkey, mode=None, other_identityKey=None, other_ratchetKey=None):
        if mode is not None:
            self.mode = mode
        else:
            if self.mode is None: # mode not selected
                sys.exit(1)

        self.conversation = self.create_conversation(other_name,
                                                     mkey,
                                                     self.mode,
                                                     self.state['DHIs_priv'],
                                                     self.state['DHIs'],
                                                     other_identityKey,
                                                     self.state['DHRs_priv'],
                                                     self.state['DHRs'],
                                                     other_ratchetKey)

        self.ratchetKey = False
        self.ratchetPKey = False

    def create_conversation(self, other_name, mkey, mode,
                            priv_identity_key, identity_key,
                            other_identity_key,
                            priv_ratchet_key=None, ratchet_key=None,
                            other_ratchet_key=None):
        if mode is ALICE_MODE:
            HKs = None
            HKr = kdf(mkey, SALTS['HK'][BOB_MODE])
            CKs = None
            CKr = kdf(mkey, SALTS['CK'][BOB_MODE])
            DHRs_priv = None
            DHRs = None
            DHRr = other_ratchet_key
            Ns = 0
            Nr = 0
            PNs = 0
            ratchet_flag = True
        else: # bob mode
            HKs = kdf(mkey, SALTS['HK'][BOB_MODE])
            HKr = None
            CKs = kdf(mkey, SALTS['CK'][BOB_MODE])
            CKr = None
            DHRs_priv = priv_ratchet_key
            DHRs = ratchet_key
            DHRr = None
            Ns = 0
            Nr = 0
            PNs = 0
            ratchet_flag = False
        RK = kdf(mkey, SALTS['RK'])
        NHKs = kdf(mkey, SALTS['NHK'][mode])
        NHKr = kdf(mkey, SALTS['NHK'][not mode])
        CONVid = kdf(mkey, SALTS['CONVid'])
        DHIr = other_identity_key

        keys = \
               { 'name': self.name,
                 'other_name': other_name,
                 'RK': RK,
                 'HKs': HKs,
                 'HKr': HKr,
                 'NHKs': NHKs,
                 'NHKr': NHKr,
                 'CKs': CKs,
                 'CKr': CKr,
                 'DHIs_priv': priv_identity_key,
                 'DHIs': identity_key,
                 'DHIr': DHIr,
                 'DHRs_priv': DHRs_priv,
                 'DHRs': DHRs,
                 'DHRr': DHRr,
                 'CONVid': CONVid,
                 'Ns': Ns,
                 'Nr': Nr,
                 'PNs': PNs,
                 'ratchet_flag': ratchet_flag,
               }

        return AxolotlConversation(self, keys, mode)

    def encrypt(self, plaintext):
        return self.conversation.encrypt(plaintext)

    def enc(self, key, plaintext):
        return encrypt_symmetric(key, plaintext)

    def dec(self, key, encrypted):
        return decrypt_symmetric(key, encrypted)

    def decrypt(self, msg):
        return self.conversation.decrypt(msg)

    def encrypt_file(self, filename):
        self.conversation.encrypt_file(filename)

    def decrypt_file(self, filename):
        self.conversation.decrypt_file(filename)

    def encrypt_pipe(self):
        self.conversation.encrypt_pipe()

    def decrypt_pipe(self):
        self.conversation.decrypt_pipe()

    def printKeys(self):
        self.conversation.print_keys()

    def saveState(self):
        self.save_conversation(self.conversation)

    def save_conversation(self, conversation):
        with self.lock:
            self.persistence.save_conversation(conversation)

    def loadState(self, name, other_name):
        self.persistence.db = self.openDB()
        self.conversation = self.load_conversation(other_name, name)
        if self.conversation:
            return
        else:
            return False

    def load_conversation(self, other_name, name=None):
        with self.lock:
            return self.persistence.load_conversation(self,
                                                      name or self.name,
                                                      other_name)

    def delete_conversation(self, conversation):
        with self.lock:
            return self.persistence.delete_conversation(conversation)

    def get_other_names(self):
        with self.lock:
            return self.persistence.get_other_names(self.name)

    def openDB(self):
        return self.persistence._open_db()

    def writeDB(self):
        self.persistence.write_db()

    def printState(self):
        self.conversation.print_state()


class AxolotlConversation:
    def __init__(self, axolotl, keys, mode, staged_hk_mk=None):
        self._axolotl = axolotl
        self.keys = keys
        self.mode = mode
        self.staged_hk_mk = staged_hk_mk or dict()
        self.staged = False

        self.handshake_key = None
        self.handshake_pkey = None

    @property
    def name(self):
        return self.keys['name']

    @name.setter
    def name(self, name):
        self.keys['name'] = name

    @property
    def other_name(self):
        return self.keys['other_name']

    @other_name.setter
    def other_name(self, other_name):
        self.keys['other_name'] = other_name

    @property
    def id_(self):
        return self.keys['CONVid']

    @id_.setter
    def id_(self, id_):
        self.keys['CONVid'] = id_

    @property
    def ns(self):
        return self.keys['Ns']

    @ns.setter
    def ns(self, ns):
        self.keys['Ns'] = ns

    @property
    def nr(self):
        return self.keys['Nr']

    @nr.setter
    def nr(self, nr):
        self.keys['Nr'] = nr

    @property
    def pns(self):
        return self.keys['PNs']

    @pns.setter
    def pns(self, pns):
        self.keys['PNs'] = pns

    @property
    def ratchet_flag(self):
        return self.keys['ratchet_flag']

    @ratchet_flag.setter
    def ratchet_flag(self, ratchet_flag):
        self.keys['ratchet_flag'] = ratchet_flag

    def _try_skipped_mk(self, msg, pad_length):
        msg1 = msg[:HEADER_LEN-pad_length]
        msg2 = msg[HEADER_LEN:]
        for skipped_mk in self.staged_hk_mk.values():
            try:
                decrypt_symmetric(skipped_mk.hk, msg1)
                body = decrypt_symmetric(skipped_mk.mk, msg2)
            except CryptoError:
                pass
            else:
                del self.staged_hk_mk[skipped_mk.mk]
                return body
        return None

    def _stage_skipped_mk(self, hkr, nr, np, ckr):
        timestamp = int(time())
        ckp = ckr
        for i in range(np - nr):
            mk = hash_(ckp + '0')
            ckp = hash_(ckp + '1')
            self.staged_hk_mk[mk] = SkippedMessageKey(mk, hkr, timestamp)
            self.staged = True
        mk = hash_(ckp + '0')
        ckp = hash_(ckp + '1')
        return ckp, mk

    def encrypt(self, plaintext):
        if self.ratchet_flag:
            self.keys['DHRs_priv'], self.keys['DHRs'] = generate_keypair()
            self.keys['HKs'] = self.keys['NHKs']
            self.keys['RK'] = hash_(self.keys['RK'] +
                                    generate_dh(self.keys['DHRs_priv'], self.keys['DHRr']))
            self.keys['NHKs'] = kdf(self.keys['RK'], SALTS['NHK'][self.mode])
            self.keys['CKs'] = kdf(self.keys['RK'], SALTS['CK'][self.mode])
            self.pns = self.ns
            self.ns = 0
            self.ratchet_flag = False
        mk = hash_(self.keys['CKs'] + '0')
        msg1 = encrypt_symmetric(
            self.keys['HKs'],
            str(self.ns).zfill(HEADER_COUNT_NUM_LEN) +
            str(self.pns).zfill(HEADER_COUNT_NUM_LEN) +
            self.keys['DHRs'])
        msg2 = encrypt_symmetric(mk, plaintext)
        pad_length = HEADER_LEN - len(msg1)
        pad = os.urandom(pad_length - HEADER_PAD_NUM_LEN) + chr(pad_length)
        msg = msg1 + pad + msg2
        self.ns += 1
        self.keys['CKs'] = hash_(self.keys['CKs'] + '1')
        return msg

    def decrypt(self, msg):
        pad = msg[HEADER_LEN-HEADER_PAD_NUM_LEN:HEADER_LEN]
        pad_length = ord(pad)
        msg1 = msg[:HEADER_LEN-pad_length]

        body = self._try_skipped_mk(msg, pad_length)
        if body and body != '':
            return body

        header = None
        if self.keys['HKr']:
            try:
                header = decrypt_symmetric(self.keys['HKr'], msg1)
            except CryptoError:
                pass
        if header and header != '':
            Np = int(header[:HEADER_COUNT_NUM_LEN])
            CKp, mk = self._stage_skipped_mk(self.keys['HKr'], self.nr, Np, self.keys['CKr'])
            try:
                body = decrypt_symmetric(mk, msg[HEADER_LEN:])
            except CryptoError:
                print 'Undecipherable message'
                sys.exit(1)
        else:
            try:
                header = decrypt_symmetric(self.keys['NHKr'], msg1)
            except CryptoError:
                pass
            if self.ratchet_flag or not header or header == '':
                print 'Undecipherable message'
                sys.exit(1)
            Np = int(header[:HEADER_COUNT_NUM_LEN])
            PNp = int(header[HEADER_COUNT_NUM_LEN:HEADER_COUNT_NUM_LEN*2])
            DHRp = header[HEADER_COUNT_NUM_LEN*2:]
            if self.keys['CKr']:
                self._stage_skipped_mk(self.keys['HKr'], self.nr, PNp, self.keys['CKr'])
            HKp = self.keys['NHKr']
            RKp = hash_(self.keys['RK'] + generate_dh(self.keys['DHRs_priv'], DHRp))
            NHKp = kdf(RKp, SALTS['NHK'][not self.mode])
            CKp = kdf(RKp, SALTS['CK'][not self.mode])
            CKp, mk = self._stage_skipped_mk(HKp, 0, Np, CKp)
            try:
                body = decrypt_symmetric(mk, msg[HEADER_LEN:])
            except CryptoError:
                pass
            if not body or body == '':
                print 'Undecipherable message'
                sys.exit(1)
            self.keys['RK'] = RKp
            self.keys['HKr'] = HKp
            self.keys['NHKr'] = NHKp
            self.keys['DHRr'] = DHRp
            self.keys['DHRs_priv'] = None
            self.keys['DHRs'] = None
            self.ratchet_flag = True
        self.nr = Np + 1
        self.keys['CKr'] = CKp
        return body

    def encrypt_file(self, filename):
        with open(filename, 'r') as f:
            plaintext = f.read()
        ciphertext = b2a(self.encrypt(plaintext)) + '\n'
        with open(filename+'.asc', 'w') as f:
            lines = [ciphertext[i:i+64] for i in xrange(0, len(ciphertext), 64)]
            for line in lines:
                f.write(line+'\n')

    def decrypt_file(self, filename):
        with open(filename, 'r') as f:
            ciphertext = a2b(f.read())
        plaintext = self.decrypt(ciphertext)
        print plaintext

    def encrypt_pipe(self):
        plaintext = sys.stdin.read()
        ciphertext = b2a(self.encrypt(plaintext)) + '\n'
        sys.stdout.write(ciphertext)
        sys.stdout.flush()

    def decrypt_pipe(self):
        ciphertext = a2b(sys.stdin.read())
        plaintext = self.decrypt(ciphertext)
        sys.stdout.write(plaintext)
        sys.stdout.flush()

    def save(self):
        self._axolotl.save_conversation(self)

    def delete(self):
        self._axolotl.delete_conversation(self)

    def print_keys(self):
        print 'Your Identity key is:\n' + b2a(self.keys['DHIs']) + '\n'
        fingerprint = hash_(self.keys['DHIs']).encode('hex').upper()
        fprint = ''
        for i in range(0, len(fingerprint), 4):
            fprint += fingerprint[i:i+2] + ':'
        print 'Your identity key fingerprint is: '
        print fprint[:-1] + '\n'
        print 'Your Ratchet key is:\n' + b2a(self.keys['DHRs']) + '\n'
        if self.handshake_key:
            print 'Your Handshake key is:\n' + b2a(self.handshake_pkey)
        else:
            print 'Your Handshake key is not available'

    def print_state(self):
        print
        print 'Warning: saving this data to disk is insecure!'
        print
        for key in sorted(self.keys):
             if 'priv' in key:
                 pass
             else:
                 if self.keys[key] is None:
                     print key + ': None'
                 elif type(self.keys[key]) is bool:
                     if self.keys[key]:
                         print key + ': True'
                     else:
                         print key + ': False'
                 elif type(self.keys[key]) is str:
                     try:
                         self.keys[key].decode('ascii')
                         print key + ': ' + self.keys[key]
                     except UnicodeDecodeError:
                         print key + ': ' + b2a(self.keys[key])
                 else:
                     print key + ': ' + str(self.keys[key])
        if self.mode is ALICE_MODE:
            print 'Mode: Alice'
        else:
            print 'Mode: Bob'


class SkippedMessageKey:
    def __init__(self, mk, hk, timestamp):
        self.mk = mk
        self.hk = hk
        self.timestamp = timestamp


class SqlitePersistence(object):
    def __init__(self, dbname, dbpassphrase, store_time, nonthreaded):
        super(SqlitePersistence, self).__init__()
        self.dbname = dbname
        self.dbpassphrase = dbpassphrase
        self.store_time = store_time
        self.nonthreaded = nonthreaded

        self.db = self._open_db()

    def _open_db(self):
        db = sqlite3.connect(':memory:', check_same_thread=self.nonthreaded)
        db.row_factory = sqlite3.Row

        with db:
            try:
                if self.dbpassphrase is not None:
                    with open(self.dbname, 'rb') as f:
                        crypt_sql = f.read()
                        try:
                            sql = decrypt_symmetric(self.dbpassphrase,
                                                    crypt_sql)
                        except CryptoError:
                            print 'Bad passphrase!'
                            sys.exit(1)
                        else:
                            db.cursor().executescript(sql)
                else:
                    with open(self.dbname, 'r') as f:
                        sql = f.read()
                        try:
                            db.cursor().executescript(sql)
                        except sqlite3.OperationalError:
                            print 'Bad sql! Password problem - cannot create the database.'
                            sys.exit(1)
            except IOError as e:
                if e.errno == errno.ENOENT:
                    self._create_db(db)
                else:
                    raise
            else:
                self._delete_expired_skipped_mk(db)
        return db

    def _create_db(self, db):
        db.execute('''
            CREATE TABLE IF NOT EXISTS
                skipped_mk (
                    my_identity,
                    to_identity,
                    HKr TEXT,
                    mk TEXT,
                    timestamp INTEGER)''')
        db.execute('''
            CREATE UNIQUE INDEX IF NOT EXISTS
                message_keys
            ON
                skipped_mk (mk)''')
        db.execute('''
            CREATE TABLE IF NOT EXISTS
                conversations (
                    my_identity TEXT,
                    other_identity TEXT,
                    RK TEXT,
                    HKs TEXT,
                    HKr TEXT,
                    NHKs TEXT,
                    NHKr TEXT,
                    CKs TEXT,
                    CKr TEXT,
                    DHIs_priv TEXT,
                    DHIs TEXT,
                    DHIr TEXT,
                    DHRs_priv TEXT,
                    DHRs TEXT,
                    DHRr TEXT,
                    CONVid TEXT,
                    Ns INTEGER,
                    Nr INTEGER,
                    PNs INTEGER,
                    ratchet_flag INTEGER,
                    mode INTEGER)''')
        db.execute('''
            CREATE UNIQUE INDEX IF NOT EXISTS
                conversation_route
            ON
                conversations (
                    my_identity,
                    other_identity)''')

    def _delete_expired_skipped_mk(self, db):
        timestamp = int(time())
        rowtime = timestamp - self.store_time
        db.execute('''
            DELETE FROM
                skipped_mk
            WHERE
                timestamp < ?''', (rowtime,))

    def _commit_skipped_mk(self, conversation):
        with self.db as db:
            db.execute('''
                DELETE FROM
                    skipped_mk
                WHERE
                    my_identity = ? AND
                    to_identity = ?''', (
                        conversation.name,
                        conversation.other_name))
            for skipped_mk in conversation.staged_hk_mk.values():
                db.execute('''
                    INSERT INTO
                        skipped_mk (
                            my_identity,
                            to_identity,
                            HKr,
                            mk,
                            timestamp)
                    VALUES (?, ?, ?, ?, ?)''', (
                        conversation.name,
                        conversation.other_name,
                        b2a(skipped_mk.hk),
                        b2a(skipped_mk.mk),
                        skipped_mk.timestamp))

    def _load_skipped_mk(self, name, other_name):
        skipped_hk_mk = dict()
        with self.db as db:
            rows = db.execute('''
                SELECT
                    *
                FROM
                    skipped_mk
                WHERE
                    my_identity = ? AND
                    to_identity = ?''', (name, other_name))
        for row in rows:
            mk = a2b(row['mk'])
            skipped_hk_mk[mk] = SkippedMessageKey(mk,
                                                  hk=a2b(row['hkr']),
                                                  timestamp=row['timestamp'])
        return skipped_hk_mk

    def write_db(self):
        with self.db as db:
            sql = bytes('\n'.join(db.iterdump()))
            if self.dbpassphrase is not None:
                crypt_sql = encrypt_symmetric(self.dbpassphrase, sql)
                with open(self.dbname, 'wb') as f:
                    f.write(crypt_sql)
            else:
                with open(self.dbname, 'w') as f:
                    f.write(sql)

    def save_conversation(self, conversation):
        HKs = 0 if conversation.keys['HKs'] is None else b2a(conversation.keys['HKs'])
        HKr = 0 if conversation.keys['HKr'] is None else b2a(conversation.keys['HKr'])
        CKs = 0 if conversation.keys['CKs'] is None else b2a(conversation.keys['CKs'])
        CKr = 0 if conversation.keys['CKr'] is None else b2a(conversation.keys['CKr'])
        DHIr = 0 if conversation.keys['DHIr'] is None else b2a(conversation.keys['DHIr'])
        DHRs_priv = 0 if conversation.keys['DHRs_priv'] is None else b2a(conversation.keys['DHRs_priv'])
        DHRs = 0 if conversation.keys['DHRs'] is None else b2a(conversation.keys['DHRs'])
        DHRr = 0 if conversation.keys['DHRr'] is None else b2a(conversation.keys['DHRr'])
        ratchet_flag = 1 if conversation.ratchet_flag else 0
        mode = 1 if conversation.mode else 0
        with self.db as db:
            db.execute('''
                REPLACE INTO
                    conversations (
                        my_identity,
                        other_identity,
                        RK,
                        HKS,
                        HKr,
                        NHKs,
                        NHKr,
                        CKs,
                        CKr,
                        DHIs_priv,
                        DHIs,
                        DHIr,
                        DHRs_priv,
                        DHRs,
                        DHRr,
                        CONVid,
                        Ns,
                        Nr,
                        PNs,
                        ratchet_flag,
                        mode)
                VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?,
                        ?, ?, ?)''', (
                    conversation.name,
                    conversation.other_name,
                    b2a(conversation.keys['RK']),
                    HKs,
                    HKr,
                    b2a(conversation.keys['NHKs']),
                    b2a(conversation.keys['NHKr']),
                    CKs,
                    CKr,
                    b2a(conversation.keys['DHIs_priv']),
                    b2a(conversation.keys['DHIs']),
                    DHIr,
                    DHRs_priv,
                    DHRs,
                    DHRr,
                    b2a(conversation.keys['CONVid']),
                    conversation.ns,
                    conversation.nr,
                    conversation.pns,
                    ratchet_flag,
                    mode))
        self._commit_skipped_mk(conversation)
        self.write_db()

    def load_conversation(self, axolotl, name, other_name):
        with self.db as db:
            cur = db.cursor()
            cur.execute('''
                SELECT
                    *
                FROM
                    conversations
                WHERE
                    my_identity = ? AND
                    other_identity = ?''', (name, other_name))
            row = cur.fetchone()
        if row:
            keys = \
                    { 'name': row['my_identity'],
                        'other_name': row['other_identity'],
                        'RK': a2b(row['rk']),
                        'NHKs': a2b(row['nhks']),
                        'NHKr': a2b(row['nhkr']),
                        'DHIs_priv': a2b(row['dhis_priv']),
                        'DHIs': a2b(row['dhis']),
                        'CONVid': a2b(row['convid']),
                        'Ns': row['ns'],
                        'Nr': row['nr'],
                        'PNs': row['pns'],
                    }
            keys['HKs'] = None if row['hks'] == '0' else a2b(row['hks'])
            keys['HKr'] = None if row['hkr'] == '0' else a2b(row['hkr'])
            keys['CKs'] = None if row['cks'] == '0' else a2b(row['cks'])
            keys['CKr'] = None if row['ckr'] == '0' else a2b(row['ckr'])
            keys['DHIr'] = None if row['dhir'] == '0' else a2b(row['dhir'])
            keys['DHRs_priv'] = None if row['dhrs_priv'] == '0' else a2b(row['dhrs_priv'])
            keys['DHRs'] = None if row['dhrs'] == '0' else a2b(row['dhrs'])
            keys['DHRr'] = None if row['dhrr'] == '0' else a2b(row['dhrr'])
            ratchet_flag = row['ratchet_flag']
            keys['ratchet_flag'] = True if ratchet_flag == 1 \
                                                else False
            mode = row['mode']
            mode = True if mode == 1 else False

            skipped_hk_mk = self._load_skipped_mk(name, other_name)

            # exit at first match
            return AxolotlConversation(axolotl, keys, mode, skipped_hk_mk)
        else:
            # if no matches
            return None

    def delete_conversation(self, conversation):
        with self.db as db:
            db.execute('''
                DELETE FROM
                    skipped_mk
                WHERE
                    to_identity = ?''', (conversation.other_name,))
            db.execute('''
                DELETE FROM
                    conversations
                WHERE
                    other_identity = ?''', (conversation.other_name,))
        self.write_db()

    def get_other_names(self, name):
        with self.db as db:
            rows = db.execute('''
                SELECT
                    other_identity
                FROM
                    conversations
                WHERE
                    my_identity = ?''', (name,))
            return [row['other_identity'] for row in rows]


def a2b(a):
    return Base64Encoder.decode(a)


def b2a(b):
    return Base64Encoder.encode(b)


def hash_(data):
    return sha256(data).decode('hex')


def kdf(secret, salt):
    return pbkdf2(secret, salt, rounds=10, prf='hmac-sha256')


Keypair = namedtuple('Keypair', 'priv pub')


def generate_keypair():
    privkey = PrivateKey.generate()
    return Keypair(bytes(privkey), bytes(privkey.public_key))


def generate_dh(a, b):
    a = PrivateKey(a)
    b = PublicKey(b)
    return bytes(Box(a, b))


def generate_3dh(a, a0, b, b0, mode=ALICE_MODE):
    if mode is ALICE_MODE:
        return hash_(generate_dh(a, b0) +
                     generate_dh(a0, b) +
                     generate_dh(a0, b0))
    else:
        return hash_(generate_dh(a0, b) +
                     generate_dh(a, b0) +
                     generate_dh(a0, b0))


def encrypt_symmetric(key, plaintext):
    nonce = nacl.utils.random(nacl.secret.SecretBox.NONCE_SIZE)
    box = nacl.secret.SecretBox(key)
    return bytes(box.encrypt(plaintext, nonce))


def decrypt_symmetric(key, ciphertext):
    box = nacl.secret.SecretBox(key)
    return box.decrypt(ciphertext)



#!/usr/bin/env python

"""
This program will encrypt or decrypt a file
"""

from pyaxo import Axolotl
import sys

your_name = raw_input('What is your name? ').strip()

# specify dbname with kwarg - default dbname is axolotl.db
# db passphrase will be prompted for - it can be specified here with dbpassprase kwarg
a = Axolotl(your_name, dbname=your_name+'.db')

try:
    if sys.argv[1] == '-e':
        other_name = \
            raw_input("What is the name of the party that you want to encrypt the file to? " ).strip()
        a.loadState(your_name, other_name)
        a.encrypt_file(sys.argv[2])
        print 'The encrypted file is: ' + sys.argv[2] + '.asc'
    else:
        other_name = \
            raw_input("What is the name of the party that you want to decrypt the file from? " ).strip()
        a.loadState(your_name, other_name)
        a.decrypt_file(sys.argv[2])
except IndexError:
    print 'Usage: ' + sys.argv[0] + ' -(e,d) <filename>'
    exit()
except KeyError:
    print 'The conversation ' + your_name + ' -> ' + other_name + \
          ' is not in the database'
    exit()

a.saveState()




#!/usr/bin/env python

from pyaxo import Axolotl
import sys

a = Axolotl('name2', dbname='name2.db', dbpassphrase=None)
a.loadState('name2', 'name1')

if sys.argv[1] == '-e':
    a.encrypt_file(sys.argv[2])
    print 'Encrypted file is ' + sys.argv[2] +'.asc'
else:
    a.decrypt_file(sys.argv[2])

a.saveState()




#!/usr/bin/env python

"""
This script will generate a pair of databases named after two conversants.
The databases can then be securely distributed to initialize axolotl.

You will need to provide your name and the other party's name. Conversations
are identified by your name and the other person's name. The conversation
should have a unique other person's name.

If you decide not to complete the initialization process, just answer no to the
question about creating a new conversation. Nothing will be saved.

If you want to reinitialize a conversation, just run the script again.
The old conversation key data will be overwritten in the databases.
"""

import sys
import binascii
from pyaxo import Axolotl

your_name = raw_input('Your name for this conversation? ').strip()
other_name = raw_input('What is the name of the other party? ').strip()
a = Axolotl(your_name,dbname=your_name+'.db')
b = Axolotl(other_name,dbname=other_name+'.db')
a.initState(other_name, b.state['DHIs'], b.handshakePKey, b.state['DHRs'], verify=False)
b.initState(your_name, a.state['DHIs'], a.handshakePKey, a.state['DHRs'], verify=False)

a.saveState()
b.saveState()
print 'The conversation ' + your_name + ' -> ' + other_name + ' has been saved in: ' + your_name + '.db'
print 'The conversation ' + other_name + ' -> ' + your_name + ' has been saved in: ' + other_name + '.db'



#!/usr/bin/env python
"""
This script is used to convert encrypted Axolotl's database files created with
pyaxo < 0.6.0 into files created with pyaxo >= 0.6.0, which replaced GPG by
PyNaCl for encryption.

To convert the databases, add each filename to the `DATABASE_NAMES` list and
run the script. It will prompt for each file's passphrase, decrypt with the
previously used cipher and encrypt with the current one. If the databases share
the same passphrase, the `dbpassphrase` declaration can be moved out of the
loop so that it will be prompted only once.
"""
import gnupg
import sys
from getpass import getpass

from pyaxo import encrypt_symmetric, hash_


DATABASE_NAMES = []


def convert_dbs():
    gpg = gnupg.GPG()
    gpg.encoding = 'utf-8'

    for dbname in DATABASE_NAMES:
        with open(dbname, 'rb') as f:
            dbpassphrase = getpass('Type passphrase for "{}": '.format(dbname))
            sql = gpg.decrypt_file(f, passphrase=dbpassphrase).data

        if sql:
            with open(dbname, 'wb') as f:
                new_crypt_sql = encrypt_symmetric(key=hash_(dbpassphrase),
                                                  plaintext=sql)
                f.write(new_crypt_sql)
        else:
            print 'Bad passphrase!'


if __name__ == '__main__':
    sys.exit(convert_dbs())



#!/usr/bin/env python

"""
This file transfer example demonstrates a couple of things:
 1) Transferring files using Axolotl to encrypt each block of the transfer
    with a different ephemeral key.
 2) Using a context manager with Axolotl.

The utility will prompt you for the location of the Axolotl key database
and the blocksize. The blocksize must be chosen so that the maximum number
of blocks is <= 255. Security is optimized by a larger number of blocks,
and transfer speed is optimized by a smaller number of blocks. If you
choose incorrectly, the utility will prompt you with a recommendation.

Key databases can be generated using e.g the init_conversation.py utility.

Syntax for receive is: ./transfer.py -r

Syntax for send is: ./transfer.py -s <filename> <target hostname or ip address>

The end of packet (EOP) and end of file (EOF) markers I use are pretty simple,
but unlikely to show up in ciphertext.
"""


from pyaxo import Axolotl
from contextlib import contextmanager
import sys
import socket
import os

try:
    location = raw_input('Database directory (default ~/.bin)? ').strip()
    if location == '': location = '~/.bin'
    location = os.path.expanduser(location)
    if sys.argv[1] == '-s':
        file_name = sys.argv[2]
        host = sys.argv[3]
        size = int(raw_input('File transfer block size? '))
    port = 50000
except IndexError:
    print 'Usage: ' + sys.argv[0] + ' -(s,r) [<filename> <host>]'
    exit()

backlog = 1

@contextmanager
def socketcontext(*args, **kwargs):
    s = socket.socket(*args, **kwargs)
    yield s
    s.close()

@contextmanager
def axo(my_name, other_name, dbname, dbpassphrase):
    a = Axolotl(my_name, dbname=dbname, dbpassphrase=dbpassphrase)
    a.loadState(my_name, other_name)
    yield a
    a.saveState()

if sys.argv[1] == '-s':
    # open socket and send data
    with socketcontext(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.connect((host, port))
        with axo('send', 'receive', dbname=location+'/send.db', dbpassphrase='1') as a:
            with open(file_name, 'rb') as f:
                plaintext = f.read()
                plainlength = len(plaintext)
                while plainlength/size > 253:
                    print 'File too large to transfer - increase size parameter'
                    print 'Recommended >= ' + str(plainlength/128) + ' bytes per block'
                    size = int(raw_input('File transfer block size? '))
                plaintext = str(len(file_name)).zfill(2) + file_name + plaintext
                while len(plaintext) > size:
                    msg = plaintext[:size]
                    if msg == '': break
                    plaintext = plaintext[size:]
                    ciphertext = a.encrypt(msg)
                    s.send(ciphertext + 'EOP')
                if len(plaintext) != 0:
                    ciphertext = a.encrypt(plaintext)
                    s.send(ciphertext + 'EOF')

            # receive confirmation
            confirmation = s.recv(1024)
            if a.decrypt(confirmation) == 'Got It!':
                print 'Transfer confirmed!'
            else:
                print 'Transfer not confirmed...'

if sys.argv[1] == '-r':
    # open socket and receive data
    with socketcontext(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        host = ''
        s.bind((host, port))
        s.listen(backlog)
        client, address = s.accept()
        with axo('receive', 'send', dbname=location+'/receive.db', dbpassphrase='1') as a:
            plaintext = ''
            ciphertext = ''
            while True:
                newtext = client.recv(1024)
                ciphertext += newtext
                if ciphertext[-3:] == 'EOF': break
            if ciphertext == '':
                print 'nothing received'
                exit()
            cipherlist = ciphertext.split('EOP')
            for item in cipherlist:
                if item[-3:] == 'EOF':
                    item = item[:-3]
                plaintext += a.decrypt(item)
            filenamelength = int(plaintext[:2])
            file_name = plaintext[2:2+filenamelength]
            with open(file_name, 'wb') as f:
                f.write(plaintext[2+filenamelength:])

            # send confirmation
            reply = a.encrypt('Got It!')
            client.send(reply)

    print file_name + ' received'



"""
Code is based on:
https://github.com/shanet/Cryptully/blob/master/src/crypto/smp.py
originally written by Shane Tully and released under the LGPL.
"""
import hashlib
import os
import random
import struct

class SMP(object):
    def __init__(self, secret=None):
        # 4096-bit safe prime (RFC 3526)
        self.mod = 0xFFFFFFFFFFFFFFFFC90FDAA22168C234C4C6628B80DC1CD129024E088A67CC74020BBEA63B139B22514A08798E3404DDEF9519B3CD3A431B302B0A6DF25F14374FE1356D6D51C245E485B576625E7EC6F44C42E9A637ED6B0BFF5CB6F406B7EDEE386BFB5A899FA5AE9F24117C4B1FE649286651ECE45B3DC2007CB8A163BF0598DA48361C55D39A69163FA8FD24CF5F83655D23DCA3AD961C62F356208552BB9ED529077096966D670C354E4ABC9804F1746C08CA18217C32905E462E36CE3BE39E772C180E86039B2783A2EC07A28FB5C55DF06F4C52C9DE2BCBF6955817183995497CEA956AE515D2261898FA051015728E5A8AAAC42DAD33170D04507A33A85521ABDF1CBA64ECFB850458DBEF0A8AEA71575D060C7DB3970F85A6E1E4C7ABF5AE8CDB0933D71E8C94E04A25619DCEE3D2261AD2EE6BF12FFA06D98A0864D87602733EC86A64521F2B18177B200CBBE117577A615D6C770988C0BAD946E208E24FA074E5AB3143DB5BFCE0FD108E4B82D120A92108011A723C12A787E6D788719A10BDBA5B2699C327186AF4E23C1A946834B6150BDA2583E9CA2AD44CE8DBBBC2DB04DE8EF92E8EFC141FBECAA6287C59474E6BC05D99B2964FA090C3A2233BA186515BE7ED1F612970CEE2D7AFB81BDD762170481CD0069127D5B05AA993B4EA988D8FDDC186FFB7DC90A6C08F4DF435C934063199FFFFFFFFFFFFFFFF
        self.modOrder = (self.mod-1) / 2
        self.gen = 2
        self.match = False

        self.secret = sha256(secret)

    def step1(self):
        self.x2 = createRandomExponent()
        self.x3 = createRandomExponent()

        self.g2 = pow(self.gen, self.x2, self.mod)
        self.g3 = pow(self.gen, self.x3, self.mod)

        (c1, d1) = self.createLogProof('1', self.x2)
        (c2, d2) = self.createLogProof('2', self.x3)

        # Send g2a, g3a, c1, d1, c2, d2
        return packList(self.g2, self.g3, c1, d1, c2, d2)

    def step2(self, buff):
        (g2a, g3a, c1, d1, c2, d2) = unpackList(buff)

        if not self.isValidArgument(g2a) or not self.isValidArgument(g3a):
            raise ValueError("Invalid g2a/g3a values")

        if not self.checkLogProof('1', g2a, c1, d1):
            raise ValueError("Proof 1 check failed")

        if not self.checkLogProof('2', g3a, c2, d2):
            raise ValueError("Proof 2 check failed")

        self.g2a = g2a
        self.g3a = g3a

        self.x2 = createRandomExponent()
        self.x3 = createRandomExponent()

        r = createRandomExponent()

        self.g2 = pow(self.gen, self.x2, self.mod)
        self.g3 = pow(self.gen, self.x3, self.mod)

        (c3, d3) = self.createLogProof('3', self.x2)
        (c4, d4) = self.createLogProof('4', self.x3)

        self.gb2 = pow(self.g2a, self.x2, self.mod)
        self.gb3 = pow(self.g3a, self.x3, self.mod)

        self.pb = pow(self.gb3, r, self.mod)
        self.qb = mulm(pow(self.gen, r, self.mod), pow(self.gb2, self.secret, self.mod), self.mod)

        (c5, d5, d6) = self.createCoordsProof('5', self.gb2, self.gb3, r)

        # Sends g2b, g3b, pb, qb, all the c's and d's
        return packList(self.g2, self.g3, self.pb, self.qb, c3, d3, c4, d4, c5, d5, d6)

    def step3(self, buff):
        (g2b, g3b, pb, qb, c3, d3, c4, d4, c5, d5, d6) = unpackList(buff)

        if not self.isValidArgument(g2b) or not self.isValidArgument(g3b) or \
           not self.isValidArgument(pb) or not self.isValidArgument(qb):
            raise ValueError("Invalid g2b/g3b/pb/qb values")

        if not self.checkLogProof('3', g2b, c3, d3):
            raise ValueError("Proof 3 check failed")

        if not self.checkLogProof('4', g3b, c4, d4):
            raise ValueError("Proof 4 check failed")

        self.g2b = g2b
        self.g3b = g3b

        self.ga2 = pow(self.g2b, self.x2, self.mod)
        self.ga3 = pow(self.g3b, self.x3, self.mod)

        if not self.checkCoordsProof('5', c5, d5, d6, self.ga2, self.ga3, pb, qb):
            raise ValueError("Proof 5 check failed")

        s = createRandomExponent()

        self.qb = qb
        self.pb = pb
        self.pa = pow(self.ga3, s, self.mod)
        self.qa = mulm(pow(self.gen, s, self.mod), pow(self.ga2, self.secret, self.mod), self.mod)

        (c6, d7, d8) = self.createCoordsProof('6', self.ga2, self.ga3, s)

        inv = self.invm(qb)
        self.ra = pow(mulm(self.qa, inv, self.mod), self.x3, self.mod)

        (c7, d9) = self.createEqualLogsProof('7', self.qa, inv, self.x3)

        # Sends pa, qa, ra, c6, d7, d8, c7, d9
        return packList(self.pa, self.qa, self.ra, c6, d7, d8, c7, d9)

    def step4(self, buff):
        (pa, qa, ra, c6, d7, d8, c7, d9) = unpackList(buff)

        if not self.isValidArgument(pa) or not self.isValidArgument(qa) or not self.isValidArgument(ra):
            raise ValueError("Invalid pa/qa/ra values")

        if not self.checkCoordsProof('6', c6, d7, d8, self.gb2, self.gb3, pa, qa):
            raise ValueError("Proof 6 check failed")

        if not self.checkEqualLogs('7', c7, d9, self.g3a, mulm(qa, self.invm(self.qb), self.mod), ra):
            raise ValueError("Proof 7 check failed")

        inv = self.invm(self.qb)
        rb = pow(mulm(qa, inv, self.mod), self.x3, self.mod)

        (c8, d10) = self.createEqualLogsProof('8', qa, inv, self.x3)

        rab = pow(ra, self.x3, self.mod)

        inv = self.invm(self.pb)
        if rab == mulm(pa, inv, self.mod):
            self.match = True

        # Send rb, c8, d10
        return packList(rb, c8, d10)

    def step5(self, buff):
        (rb, c8, d10) = unpackList(buff)

        if not self.isValidArgument(rb):
            raise ValueError("Invalid rb values")

        if not self.checkEqualLogs('8', c8, d10, self.g3b, mulm(self.qa, self.invm(self.qb), self.mod), rb):
            raise ValueError("Proof 8 check failed")

        rab = pow(rb, self.x3, self.mod)

        inv = self.invm(self.pb)
        if rab == mulm(self.pa, inv, self.mod):
            self.match = True

    def createLogProof(self, version, x):
        randExponent = createRandomExponent()
        c = sha256(version + str(pow(self.gen, randExponent, self.mod)))
        d = (randExponent - mulm(x, c, self.modOrder)) % self.modOrder
        return (c, d)

    def checkLogProof(self, version, g, c, d):
        gd = pow(self.gen, d, self.mod)
        gc = pow(g, c, self.mod)
        gdgc = gd * gc % self.mod
        return (sha256(version + str(gdgc)) == c)

    def createCoordsProof(self, version, g2, g3, r):
        r1 = createRandomExponent()
        r2 = createRandomExponent()

        tmp1 = pow(g3, r1, self.mod)
        tmp2 = mulm(pow(self.gen, r1, self.mod), pow(g2, r2, self.mod), self.mod)

        c = sha256(version + str(tmp1) + str(tmp2))

        # TODO: make a subm function
        d1 = (r1 - mulm(r, c, self.modOrder)) % self.modOrder
        d2 = (r2 - mulm(self.secret, c, self.modOrder)) % self.modOrder

        return (c, d1, d2)

    def checkCoordsProof(self, version, c, d1, d2, g2, g3, p, q):
        tmp1 = mulm(pow(g3, d1, self.mod), pow(p, c, self.mod), self.mod)

        tmp2 = mulm(mulm(pow(self.gen, d1, self.mod), pow(g2, d2, self.mod), self.mod), pow(q, c, self.mod), self.mod)

        cprime = sha256(version + str(tmp1) + str(tmp2))

        return (c == cprime)

    def createEqualLogsProof(self, version, qa, qb, x):
        r = createRandomExponent()
        tmp1 = pow(self.gen, r, self.mod)
        qab = mulm(qa, qb, self.mod)
        tmp2 = pow(qab, r, self.mod)

        c = sha256(version + str(tmp1) + str(tmp2))
        tmp1 = mulm(x, c, self.modOrder)
        d = (r - tmp1) % self.modOrder

        return (c, d)

    def checkEqualLogs(self, version, c, d, g3, qab, r):
        tmp1 = mulm(pow(self.gen, d, self.mod), pow(g3, c, self.mod), self.mod)

        tmp2 = mulm(pow(qab, d, self.mod), pow(r, c, self.mod), self.mod)

        cprime = sha256(version + str(tmp1) + str(tmp2))
        return (c == cprime)

    def invm(self, x):
        return pow(x, self.mod-2, self.mod)

    def isValidArgument(self, val):
        return (val >= 2 and val <= self.mod-2)

def packList(*items):
    buff = ''

    # For each item in the list, convert it to a byte string and add its length as a prefix
    for item in items:
        bytes = longToBytes(item)
        buff += struct.pack('!I', len(bytes)) + bytes

    return buff

def unpackList(buff):
    items = []

    index = 0
    while index < len(buff):
        # Get the length of the long (4 byte int before the actual long)
        length = struct.unpack('!I', buff[index:index+4])[0]
        index += 4

        # Convert the data back to a long and add it to the list
        item = bytesToLong(buff[index:index+length])
        items.append(item)
        index += length

    return items

def bytesToLong(bytes):
    length = len(bytes)
    string = 0
    for i in range(length):
        string += byteToLong(bytes[i:i+1]) << 8*(length-i-1)
    return string

def longToBytes(long):
    bytes = ''
    while long != 0:
        bytes = longToByte(long & 0xff) + bytes
        long >>= 8
    return bytes

def byteToLong(byte):
    return struct.unpack('B', byte)[0]

def longToByte(long):
    return struct.pack('B', long)

def mulm(x, y, mod):
    return x * y % mod

def createRandomExponent():
    return random.getrandbits(512*8)

def sha256(message):
    return long(hashlib.sha256(str(message)).hexdigest(), 16)



#!/usr/bin/env python

import copy
import os
from pyaxo import Axolotl

name1 = 'Angie'
name2 = 'Barb'

a = Axolotl(name1, dbname='name1.db', dbpassphrase=None)
b = Axolotl(name2, dbname='name2.db', dbpassphrase=None)

a.loadState(name1, name2)
b.loadState(name2, name1)

topic = ['   My Name',
          'Other Name',
          '        RK',
          '       HKs',
          '       HKr',
          '      NHKs',
          '      NHKr',
          '       CKs',
          '       CKr',
          ' DHIs_priv',
          '      DHIs',
          '      DHIr',
          ' DHRs_priv',
          '      DHRs',
          '      DHRr',
          '    CONVid',
          '        Ns',
          '        Nr',
          '       PNs',
          '   ratchet',
          '      mode']

def hilite(text, c=False):
    attr = []
    if c:
        attr.append('41')
    return '\x1b[%sm%s\x1b[0m' % (';'.join(attr), text)

data_old = False
os.system('clear')
while True:
    print '\x1b[;32m     Name: 1 2\x1b[0m'
    print '--------------'
    a.loadState(name1, name2)
    b.loadState(name2, name1)
    databases = (a.db, b.db)
    data = []
    a_chg = False
    b_chg = False
    for number, database in enumerate(databases):
        cur = database.cursor()
        cur.execute('SELECT * from conversations')
        data += [cur.fetchall()]
    if not data_old:
        data_old = data
    for i in range(len(data[0][0])):
        if data[0][0][i] != data_old[0][0][i]: a_chg=True
        if data[1][0][i] != data_old[1][0][i]: b_chg=True
        if topic[i] == '      mode':
            if data[0][0][i] == 1:
                var = 'A'
                var2 = 'B'
            else:
                var = 'B'
                var2 = 'A'
        elif topic[i]=='        Ns' or topic[i]=='        Nr' or topic[i]=='       PNs':
            var = data[0][0][i]
            var2 = data[1][0][i]
        elif topic[i] == '   ratchet':
            var = 'F'
            var2 = 'F'
            if data[0][0][i] == 1:
                var = 'T'
            elif data[1][0][i] == 1:
                var2 = 'T'
        else:
            var = '*'
            var2 = '*'
        print topic[i], hilite(var, a_chg), hilite(var2, b_chg)
        a_chg = False
        b_chg = False
    print '--------------'
    ans = raw_input('Load new state? ')
    if ans=='q' or ans=='n': exit()
    os.system('clear')
    data_old = data



def test_init_state(axolotl_a, axolotl_b, exchange):
    axolotl_a.initState(other_name=axolotl_b.name,
                        other_identityKey=axolotl_b.state['DHIs'],
                        other_handshakeKey=axolotl_b.handshakePKey,
                        other_ratchetKey=axolotl_b.state['DHRs'],
                        verify=False)

    axolotl_b.initState(other_name=axolotl_a.name,
                        other_identityKey=axolotl_a.state['DHIs'],
                        other_handshakeKey=axolotl_a.handshakePKey,
                        other_ratchetKey=axolotl_a.state['DHRs'],
                        verify=False)

    exchange(axolotl_a, axolotl_b)


def test_create_state(axolotl_a, axolotl_b, exchange):
    mkey = 'masterkey'

    axolotl_a.createState(other_name=axolotl_b.name,
                          mkey=mkey,
                          mode=True,
                          other_ratchetKey=axolotl_b.state['DHRs'])
    axolotl_b.createState(other_name=axolotl_a.name,
                          mkey=mkey,
                          mode=False)

    exchange(axolotl_a, axolotl_b)



#!/usr/bin/env python

from pyaxo import Axolotl
import sys
import os

# start with a fresh database
try:
    os.remove('./name1.db')
    os.remove('./name2.db')
except OSError:
    pass

# unencrypted databases
a = Axolotl('name1', dbname='name1.db', dbpassphrase=None)
b = Axolotl('name2', dbname='name2.db', dbpassphrase=None)

a.initState('name2', b.state['DHIs'], b.handshakePKey, b.state['DHRs'], verify=False)
b.initState('name1', a.state['DHIs'], a.handshakePKey, a.state['DHRs'], verify=False)

a.saveState()
b.saveState()



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import versioneer
from glob import glob


BASE_DIRECTORY = '/usr/share/pyaxo'

setup(
    name='pyaxo',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='Python implementation of the Axolotl ratchet protocol',
    author='David R. Andersen',
    author_email='k0rx@RXcomm.net',
    url='https://github.com/rxcomm/pyaxo',
    py_modules=[
        'pyaxo'
    ],
    install_requires=[
        'passlib>=1.6.1',
        'pynacl>=1.0.1',
    ],
    data_files=[
        (BASE_DIRECTORY + '/examples', glob('examples/*')),
        (BASE_DIRECTORY + '/tests', glob('tests/*')),
        (BASE_DIRECTORY + '/utilities', glob('utilities/*')),
        (BASE_DIRECTORY, ['COPYING']),
        (BASE_DIRECTORY, ['README.rst']),
    ],
)



import sqlite3
from copy import deepcopy

import pytest

from pyaxo import Axolotl

from . import utils


PASSPHRASES = [None, '123', '321']


class TestDefaultDatabase:
    dbs = [utils.DEFAULT_DB]

    def test_shared_db(self):
        # create two instance classes - one which will share its database
        # (note that Dick and Harry's passphrases must match or Harry won't
        # be able to load Dick's saved database)
        shared_pass = 'shared passphrase'
        tom = Axolotl('Tom', dbpassphrase="tom's passphrase")
        dick = Axolotl('Dick', dbpassphrase=shared_pass)

        # initialize Tom and Dick's states
        tom.initState(other_name=dick.name,
                      other_identityKey=dick.state['DHIs'],
                      other_handshakeKey=dick.handshakePKey,
                      other_ratchetKey=dick.state['DHRs'],
                      verify=False)
        dick.initState(other_name=tom.name,
                       other_identityKey=tom.state['DHIs'],
                       other_handshakeKey=tom.handshakePKey,
                       other_ratchetKey=tom.state['DHRs'],
                       verify=False)

        # get the plaintext
        msg = 'plaintext'

        # Tom encrypts it to Dick
        ciphertext = tom.encrypt(msg)

        # save Dick's state prior to decrypting the message
        dick.saveState()

        # Dick decrypts the ciphertext
        assert dick.decrypt(ciphertext) == msg

        # now load Dick's state to Harry
        harry = Axolotl('Harry', dbpassphrase=shared_pass)
        harry.loadState(dick.name, tom.name)

        # Harry decrypts the ciphertext
        assert harry.decrypt(ciphertext) == msg

    @pytest.mark.parametrize('passphrase_1', PASSPHRASES)
    @pytest.mark.parametrize('passphrase_0', PASSPHRASES)
    def test_passphrase(self, passphrase_0, passphrase_1):
        a = Axolotl('Angie', dbpassphrase=passphrase_0)
        b = Axolotl('Barb', dbpassphrase=None)

        a.initState(other_name=b.name,
                    other_identityKey=b.state['DHIs'],
                    other_handshakeKey=b.handshakePKey,
                    other_ratchetKey=b.state['DHRs'],
                    verify=False)
        a.saveState()

        if passphrase_0 == passphrase_1:
            a = Axolotl('Angie', dbpassphrase=passphrase_1)
            assert isinstance(a.db, sqlite3.Connection)
        else:
            with pytest.raises(SystemExit):
                a = Axolotl('Angie', dbpassphrase=passphrase_1)

    def test_delete_conversation(
            self, axolotl_a, axolotl_b, axolotl_c,
            a_identity_keys, b_identity_keys, c_identity_keys,
            a_handshake_keys, b_handshake_keys, c_handshake_keys,
            a_ratchet_keys, b_ratchet_keys, c_ratchet_keys):
        conv_b = axolotl_a.init_conversation(
            axolotl_b.name,
            priv_identity_key=a_identity_keys.priv,
            identity_key=a_identity_keys.pub,
            priv_handshake_key=a_handshake_keys.priv,
            other_identity_key=b_identity_keys.pub,
            other_handshake_key=b_handshake_keys.pub,
            priv_ratchet_key=a_ratchet_keys.priv,
            ratchet_key=a_ratchet_keys.pub,
            other_ratchet_key=b_ratchet_keys.pub)

        conv_b.save()
        conv_b.delete()

        assert not axolotl_a.load_conversation(axolotl_b.name)

    def test_get_other_names(
            self, axolotl_a, axolotl_b, axolotl_c,
            a_identity_keys, b_identity_keys, c_identity_keys,
            a_handshake_keys, b_handshake_keys, c_handshake_keys,
            a_ratchet_keys, b_ratchet_keys, c_ratchet_keys):
        conv_b = axolotl_a.init_conversation(
            axolotl_b.name,
            priv_identity_key=a_identity_keys.priv,
            identity_key=a_identity_keys.pub,
            priv_handshake_key=a_handshake_keys.priv,
            other_identity_key=b_identity_keys.pub,
            other_handshake_key=b_handshake_keys.pub,
            priv_ratchet_key=a_ratchet_keys.priv,
            ratchet_key=a_ratchet_keys.pub,
            other_ratchet_key=b_ratchet_keys.pub)

        conv_c = axolotl_a.init_conversation(
            axolotl_c.name,
            priv_identity_key=a_identity_keys.priv,
            identity_key=a_identity_keys.pub,
            priv_handshake_key=a_handshake_keys.priv,
            other_identity_key=c_identity_keys.pub,
            other_handshake_key=c_handshake_keys.pub,
            priv_ratchet_key=a_ratchet_keys.priv,
            ratchet_key=a_ratchet_keys.pub,
            other_ratchet_key=c_ratchet_keys.pub)

        conv_b.save()
        conv_c.save()

        assert (sorted(axolotl_a.get_other_names()) ==
                sorted([axolotl_b.name, axolotl_c.name]))


class TestIndividualDatabases:
    dbs = ['angie.db', 'barb.db']

    def test_individual_dbs(self, exchange):
        # create two instance classes with encrypted databases
        a = Axolotl('angie', dbname=self.dbs[0], dbpassphrase=self.dbs[0])
        b = Axolotl('barb', dbname=self.dbs[1], dbpassphrase=self.dbs[1])

        # initialize the states
        a.initState(other_name=b.name,
                    other_identityKey=b.state['DHIs'],
                    other_handshakeKey=b.handshakePKey,
                    other_ratchetKey=b.state['DHRs'],
                    verify=False)
        b.initState(other_name=a.name,
                    other_identityKey=a.state['DHIs'],
                    other_handshakeKey=a.handshakePKey,
                    other_ratchetKey=a.state['DHRs'],
                    verify=False)

        # save the states
        a.saveState()
        b.saveState()

        # reload the databases
        a = Axolotl('angie', dbname=self.dbs[0], dbpassphrase=self.dbs[0])
        b = Axolotl('barb', dbname=self.dbs[1], dbpassphrase=self.dbs[1])

        # load their states
        a.loadState(a.name, b.name)
        b.loadState(b.name, a.name)

        # send some messages back and forth
        exchange(a, b)

    def test_persist_skipped_mk(
            self, a_identity_keys, a_handshake_keys, a_ratchet_keys,
            b_identity_keys, b_handshake_keys, b_ratchet_keys):
        a = Axolotl('angie', dbname=self.dbs[0], dbpassphrase=self.dbs[0])
        b = Axolotl('barb', dbname=self.dbs[1], dbpassphrase=self.dbs[1])

        conv_a = a.init_conversation(
            b.name,
            priv_identity_key=a_identity_keys.priv,
            identity_key=a_identity_keys.pub,
            priv_handshake_key=a_handshake_keys.priv,
            other_identity_key=b_identity_keys.pub,
            other_handshake_key=b_handshake_keys.pub,
            priv_ratchet_key=a_ratchet_keys.priv,
            ratchet_key=a_ratchet_keys.pub,
            other_ratchet_key=b_ratchet_keys.pub)

        conv_b = b.init_conversation(
            a.name,
            priv_identity_key=b_identity_keys.priv,
            identity_key=b_identity_keys.pub,
            priv_handshake_key=b_handshake_keys.priv,
            other_identity_key=a_identity_keys.pub,
            other_handshake_key=a_handshake_keys.pub,
            priv_ratchet_key=b_ratchet_keys.priv,
            ratchet_key=b_ratchet_keys.pub,
            other_ratchet_key=a_ratchet_keys.pub)

        pt = [utils.PLAINTEXT.format(i) for i in range(5)]
        ct = list()

        utils.encrypt(conv_a, 0, pt, ct)
        utils.decrypt(conv_b, 0, pt, ct)
        utils.encrypt(conv_a, 1, pt, ct)
        utils.encrypt(conv_a, 2, pt, ct)
        utils.decrypt(conv_b, 2, pt, ct)
        utils.encrypt(conv_a, 3, pt, ct)
        utils.encrypt(conv_a, 4, pt, ct)
        utils.decrypt(conv_b, 4, pt, ct)

        # make sure there are staged skipped keys
        assert conv_b.staged_hk_mk

        # save the database, copy the staged keys dict and delete the objects
        conv_b.save()
        persisted_hk_mk = deepcopy(conv_b.staged_hk_mk)
        del b, conv_b

        # load the conversation from disk
        B = Axolotl('barb', dbname=self.dbs[1], dbpassphrase=self.dbs[1])
        conv_B = B.load_conversation(a.name)

        # assert both dicts have the same content
        assert conv_B.staged_hk_mk.keys() == persisted_hk_mk.keys()
        for mk in conv_B.staged_hk_mk:
            assert conv_B.staged_hk_mk[mk].mk == persisted_hk_mk[mk].mk
            assert conv_B.staged_hk_mk[mk].hk == persisted_hk_mk[mk].hk
            assert (conv_B.staged_hk_mk[mk].timestamp ==
                    persisted_hk_mk[mk].timestamp)

        # decrypt the skipped messages
        utils.decrypt(conv_B, 1, pt, ct)
        utils.decrypt(conv_B, 3, pt, ct)


@pytest.fixture(autouse=True)
def setup_teardown_dbs(request):
    remove_dbs(request.cls.dbs)
    yield
    remove_dbs(request.cls.dbs)


def remove_dbs(dbs):
    for db in dbs:
        utils.remove_db(db)
